/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _INTERIM_
#define _INTERIM_

#include "interim.hpp"
#include <unistd.h>

interim::pos_t::pos_t() {
	y = 0;
	x = 0;
}

interim::interim() {
	state = 0;
	rows = 0;
	topstart();
	automove();
}

void interim::topstart(bool what) {
	if (what)
		state |= TOPSTART;
	else
		state &= ~TOPSTART;
}

void interim::automove(bool what) {
	if (what)
		state |= AUTOMOVE | LINEDOWN;
	else
		state &= ~(AUTOMOVE | LINEDOWN);
}

void interim::open() {
	in.assign(STDIN_FILENO);
	in.sleep(true);
	in.readability(true);
	initscr();
	nodelay(stdscr, TRUE);
	keypad(stdscr, TRUE);
	noecho();
	curs_set(0);
	keys.reserve(10);
	key = keys.begin();
	left = 0;
	dimensions();
	top = win.pos.y;
	set_escdelay(25);
}

void interim::close() {
	endwin();
}

void interim::dimensions() {
	getmaxyx(stdscr, scr.max.y, scr.max.x);
	win.max.y = lines.size();
	win.max.x = 0;
	for (lines_t::iterator line = lines.begin(); line != lines.end(); ++line) {
		if (line->size() > win.max.x)
			win.max.x = line->size();
	}
}

interim::keys_t interim::keystroke() {
	size_t left = in.hasToRead();
	keys_t::iterator end = key+this->left-left;
	keys_t ks(key, end);
	key = end;
	this->left = left;
	if (!left) {
		in.fetch(&keys[0], keys.capacity());
		this->left = keys.capacity();
		key = keys.begin();
	}
	return ks;
}

interim::nkeys_t interim::parseKeys() {
	nkeys_t nkeys;
	int nkey;
	while ((nkey = getch()) != ERR) {
		move(nkey);
		nkeys.push_back(nkey);
	}
	return nkeys;
}

void interim::move(int nkey) {
	if (win.max.x > scr.max.x) {
		if (scr.pos.x > win.max.x-scr.max.x)
			scr.pos.x = win.max.x-scr.max.x;
		switch (nkey) {
			case KEY_LEFT:
				if (scr.pos.x)
					--scr.pos.x;
			break;
			case KEY_RIGHT:
				if (scr.pos.x < win.max.x-scr.max.x)
					++scr.pos.x;
			break;
		}
	}
	else
		scr.pos.x = 0;
	if (win.max.y > scr.max.y) {
		if (scr.pos.y > win.max.y-scr.max.y)
			scr.pos.y = win.max.y-scr.max.y;
		switch (nkey) {
			case KEY_UP:
				if (scr.pos.y)
					--scr.pos.y;
				if (state & AUTOMOVE)
					state &= ~LINEDOWN;
			break;
			case KEY_DOWN:
				if (scr.pos.y < win.max.y-scr.max.y)
					++scr.pos.y;
				if (state & AUTOMOVE and scr.pos.y == win.max.y-scr.max.y)
					state |= LINEDOWN;
			break;	
		}
	}
	else
		scr.pos.y = 0;
}

void interim::update(bool force) {
	dimensions();
	if (state & LINEDOWN and win.max.y > scr.max.y)
		scr.pos.y = win.max.y-scr.max.y;
	if (!(state & TOPSTART) and scr.max.y > win.max.y)
		win.pos.y = scr.max.y-win.max.y;
	else
		win.pos.y = 0;
	if (win.pos.y > top) {
		for (size_t y = 0; y < win.pos.y-top; ++y)
			mvprintw(top+y, 0, "\n");
	}
	top = win.pos.y;
	for (lines_t::iterator line = lines.begin()+scr.pos.y; line != lines.end() and line != lines.begin()+scr.pos.y+scr.max.y; ++line) {
		std::string row;
		if (scr.pos.x < line->size())
			row = line->substr(scr.pos.x, win.max.x > scr.max.x ? scr.max.x : win.max.x);
		if (row.size() < scr.max.x)
			row += "\n";
		mvprintw(win.pos.y++, 0, row.c_str());
		/* Bottom line of win sometimes hangs when scrolling.
		 * If win.max.y > scr.max.y and we don't have the bottom line of win at the bottom line of scr,
		 * then the bottom line of scr does not hang. In xterm nothing hangs. Very strange...
		 * A clear() at the beginning of this method would solve it but it always puts blank chars
		 * making it very unefficient. Only hangs when scr is very small!
		 */
	}
	if (win.max.y < rows) {
		for (size_t y = 0; y < rows-win.max.y; ++y)
			mvprintw(win.pos.y+y, 0, "\n");
	}
	rows = win.max.y;
	refresh();
}

interim::~interim() {
	close();
}

#endif
