/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PIBOX_FIXED_H
#define _PIBOX_FIXED_H

#include <sys/types.h>
#include <string>
#include <sstream>

#define PIBOX_ENABLE_ITERATORS 1
#define PIBOX_ENABLE_SAFE_MODE 1

namespace pibox {
	template <typename data_t>
	class fixed;
	template <typename number_t>
	inline std::string to_string(number_t number) {
		return ((std::ostringstream&)(std::ostringstream() << number)).str();
	}
}

template <typename data_t>
class pibox::fixed {
	size_t data_size;
	data_t* data_front;
	data_t* data_back;
	data_t* ram_begin;
	data_t* ram_end;
public:
	inline void increment(data_t*& data) {
		++data;
		if (data == ram_end)
			data = ram_begin;
	}
	inline data_t* increment(data_t*const& data, size_t size) {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!(size < data_size+PIBOX_ENABLE_ITERATORS))
			throw (std::string)"Would increment out of range";
	#endif
		size_t more = ram_end-data;
		if (size < more)
			return data+size;
		else
			return ram_begin+size-more;
	}
	inline void decrement(data_t*& data) {
		if (data == ram_begin and data_size)
			data = ram_end-1;
		else
			--data;
	}
	inline data_t* decrement(data_t*const& data, size_t size) {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!(size < data_size+PIBOX_ENABLE_ITERATORS))
			throw (std::string)"Would decrement out of range";
	#endif
		size_t less = data-ram_begin;
		if (less < size)
			return ram_end-size+less;
		else
			return data-size;
	}
	inline size_t difference(const data_t*const& minuend, const data_t*const& subtrahent) {
		return subtrahent > minuend ? ram_end-ram_begin-(subtrahent-minuend) : minuend-subtrahent;
	}
	class iterator {
		friend fixed;
		data_t* data;
		fixed<data_t>* owner;
		iterator(data_t* data, fixed<data_t>* owner) {
			this->data = data;
			this->owner = owner;
		}
	public:
		iterator() {
			data = 0;
			owner = 0;
		}
		iterator(fixed<data_t>& item) {
			data = item.data;
			owner = &item;
		}
		virtual ~iterator() {
		}
		data_t& operator*() {
			return *data;
		}
		data_t* operator->() {
			return data;
		}
		iterator& operator++() {
			owner->increment(data);
			return *this;
		}
		iterator operator++(int) {
			iterator that(*this);
			owner->increment(data);
			return that;
		}
		iterator& operator+=(const iterator& that) {
			return *this = iterator(owner->increment(data, that.owner->difference(that.data, that.owner->data_front)), owner);
		}
		iterator& operator+=(const size_t& size) {
			return *this = iterator(owner->increment(data, size), owner);
		}
		iterator operator+(const iterator& that) const {
			return iterator(owner->increment(data, that.owner->difference(that.data, that.owner->data_front)), owner);
		}
		iterator operator+(const size_t& size) const {
			return iterator(owner->increment(data, size), owner);
		}
		iterator& operator--()  {
			owner->decrement(data);
			return *this;
		}
		iterator operator--(int) {
			iterator that(*this);
			owner->decrement(data);
			return that;
		}
		iterator& operator-=(const iterator& that) {
			return *this = iterator(owner->decrement(data, that.owner->difference(that.data, that.owner->data_front)), owner);
		}
		iterator& operator-=(const size_t& size) {
			return *this = iterator(owner->decrement(data, size), owner);
		}
		iterator operator-(const iterator& that) const {
			return iterator(owner->decrement(data, that.owner->difference(that.data, that.owner->data_front)), owner);
		}
		iterator operator-(const size_t& size) const {
			return iterator(owner->decrement(data, size), owner);
		}
		bool operator==(const iterator& that) {
			return data == that.data;
		}
		bool operator!=(const iterator& that) {
			return data != that.data;
		}
		bool operator<(const iterator& that) {
			return owner->difference(data, owner->data_front) < that.owner->difference(that.data, that.owner->data_front);
		}
		bool operator<=(const iterator& that) {
			return owner->difference(data, owner->data_front) <= that.owner->difference(that.data, that.owner->data_front);
		}
		bool operator>(const iterator& that) {
			return owner->difference(data, owner->data_front) > that.owner->difference(that.data, that.owner->data_front);
		}
		bool operator>=(const iterator& that) {
			return owner->difference(data, owner->data_front) >= that.owner->difference(that.data, that.owner->data_front);
		}
	};
	class reverse_iterator {
		friend fixed;
		data_t* data;
		fixed<data_t>* owner;
		reverse_iterator(data_t* data, fixed<data_t>* owner) {
			this->data = data;
			this->owner = owner;
		}
	public:
		reverse_iterator() {
			data = 0;
			owner = 0;
		}
		~reverse_iterator() {
		}
		data_t& operator*() {
			return *data;
		}
		data_t* operator->() {
			return data;
		}
		reverse_iterator& operator++() {
			owner->decrement(data);
			return *this;
		}
		reverse_iterator operator++(int) {
			reverse_iterator that(*this);
			owner->decrement(data);
			return that;
		}
		reverse_iterator& operator+=(const reverse_iterator& that) {
			return *this = reverse_iterator(owner->decrement(data, that.owner->difference(that.owner->data_back, that.data)), owner);
		}
		reverse_iterator& operator+=(const size_t& size) {
			return *this = reverse_iterator(owner->decrement(data, size), owner);
		}
		reverse_iterator operator+(const reverse_iterator& that) const {
			return reverse_iterator(owner->decrement(data, that.owner->difference(that.owner->data_back, that.data)), owner);
		}
		reverse_iterator operator+(const size_t& size) const {
			return reverse_iterator(owner->decrement(data, size), owner);
		}
		reverse_iterator& operator--()  {
			owner->increment(data);
			return *this;
		}
		reverse_iterator operator--(int) {
			reverse_iterator that(*this);
			owner->increment(data);
			return that;
		}
		reverse_iterator& operator-=(const reverse_iterator& that) {
			return *this = reverse_iterator(owner->increment(data, that.owner->difference(that.owner->data_back, that.data)), owner);
		}
		reverse_iterator& operator-=(const size_t& size) {
			return *this = reverse_iterator(owner->increment(data, size), owner);
		}
		reverse_iterator operator-(const reverse_iterator& that) const {
			return reverse_iterator(owner->increment(data, that.owner->difference(that.owner->data_back, that.data)), owner);
		}
		reverse_iterator operator-(const size_t& size) const {
			return reverse_iterator(owner->increment(data, size), owner);
		}
		bool operator==(const reverse_iterator& that) {
			return data == that.data;
		}
		bool operator!=(const reverse_iterator& that) {
			return data != that.data;
		}
		bool operator<(const reverse_iterator& that) {
			return owner->difference(owner->data_back, data) < that.owner->difference(that.owner->data_back, that.data);
		}
		bool operator<=(const reverse_iterator& that) {
			return owner->difference(owner->data_back, data) <= that.owner->difference(that.owner->data_back, that.data);
		}
		bool operator>(const reverse_iterator& that) {
			return owner->difference(owner->data_back, data) > that.owner->difference(that.owner->data_back, that.data);
		}
		bool operator>=(const reverse_iterator& that) {
			return owner->difference(owner->data_back, data) >= that.owner->difference(that.owner->data_back, that.data);
		}
	};
	iterator begin() {
		return iterator(data_front, this);
	}
	iterator end() {
		return ++iterator(data_back, this);
	}
	reverse_iterator rbegin() {
		return reverse_iterator(data_back, this);
	}
	reverse_iterator rend() {
		return ++reverse_iterator(data_front, this);
	}
	inline size_t difference(iterator const& minuend, iterator const& subtrahent) {
		return difference(minuend.data, subtrahent.data);
	}
	inline size_t difference(reverse_iterator const& minuend, reverse_iterator const& subtrahent) {
		return difference(subtrahent.data, minuend.data);
	}
	inline bool continious(const data_t*const& front, const data_t*const& back) {
		return front <= back;
	}
	inline bool continious(iterator front, iterator back) {
		return continious(&*front, &*back);
	}
	inline bool continious(reverse_iterator rback, reverse_iterator rend) {
		return continious(&*rend, &*rback);
	}
	inline bool continious_front(size_t size) {
		return continious(data_front, increment(data_front, size));
	}
	inline bool continious_front(size_t size, size_t front) {
		data_t* data = increment(data_front, front);
		return continious(data, increment(data, size));
	}
	inline bool continious_back(size_t size) {
		return continious(decrement(data_back, size), data_back);
	}
	inline bool continious_back(size_t size, size_t back) {
		data_t* data = decrement(data_back, back);
		return continious(decrement(data, size), data);
	}
	inline bool continious(ssize_t size) {
		if (size < 0)
			return continious_back(-size);
		else
			return continious_front(size);
	}
	inline bool continious(ssize_t size, size_t item) {
		if (size < 0)
			return continious_back(-size, item);
		else
			return continious_front(size, item);
	}
	fixed() {
		ram_begin = 0;
		reserve(0);
	}
	fixed(size_t size) {
		ram_begin = 0;
		data_size = 0;
		reserve(size);
	}
	~fixed() {
		delete[] ram_begin;
	}
	void reserve(size_t size) {
		if (!size) {
			free();
		}
		else {
			data_t* new_ram_begin = 0;
			try {
				new_ram_begin = new data_t[size+PIBOX_ENABLE_ITERATORS];
			}
			catch (std::bad_alloc& error) {
				throw (std::string)"No RAM left to reserve "+to_string(size+PIBOX_ENABLE_ITERATORS)+" B";
			}
			if (data_size > size)
				data_size = size;
			for (size_t i = 0; i < data_size; ++i, increment(data_front))
				new_ram_begin[i] = *data_front;
			delete[] ram_begin;
			ram_begin = new_ram_begin;
			ram_end = ram_begin+size+PIBOX_ENABLE_ITERATORS;
			data_front = ram_begin;
			data_back = ram_begin+data_size-1;
		}
	}
	void push_front(data_t data) {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!left())
			throw (std::string)"No front left to push";
	#endif
		decrement(data_front);
		*data_front = data;
		++data_size;
	}
	void push_back(data_t data) {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!left())
			throw (std::string)"No back left to push";
	#endif
		increment(data_back);
		*data_back = data;
		++data_size;
	}
	inline data_t& front() {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!used())
			throw (std::string)"No front";
	#endif
		return *data_front;
	}
	inline data_t& front(size_t size) {
		return *increment(data_front, size);
	}
	inline data_t& back() {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!used())
			throw (std::string)"No back";
	#endif
		return *data_back;
	}
	inline data_t& back(size_t size) {
		return *decrement(data_back, size);
	}
	inline void pop_front() {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!used())
			throw (std::string)"No front to pop";
	#endif
		increment(data_front);
		--data_size;
	}
	inline void pop_back() {
	#if PIBOX_ENABLE_SAFE_MODE
		if (!used())
			throw (std::string)"No back to pop";
	#endif
		decrement(data_back);
		--data_size;
	}
	inline void clear() {
		data_size = 0;
		data_front = ram_begin;
		data_back = ram_begin-1;
	}
	inline void free() {
		delete[] ram_begin;
		ram_begin = 0;
		ram_end = 0;
		clear();
	}
	inline size_t size() {
		return ram_end-PIBOX_ENABLE_ITERATORS-ram_begin;
	}
	inline size_t used() {
		return data_size;
	}
	inline size_t left() {
		return size()-used();
	}
	inline size_t left_front() {
		if (continious(data_front, data_back)) {
			size_t less = data_front-ram_begin;
			size_t more = left();
			if (less > more)
				less = more;
			return less;
		}
		else
			return data_front-data_back-PIBOX_ENABLE_ITERATORS;
	}
	inline size_t left_back() {
		if (continious(data_front, data_back)) {
			size_t less = ram_end-1-data_back;
			size_t more = left();
			if (less > more)
				less = more;
			return less;
		}
		else
			return data_front-data_back-PIBOX_ENABLE_ITERATORS;
	}
	inline data_t& operator[](const ssize_t& size) {
		if (size < 0)
			return back(-(1+size));
		else
			return front(size);
	}
};

#endif
