/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PIBOX_MIXED_H
#define _PIBOX_MIXED_H

#include "fixed.hpp"

namespace pibox {
	class mixed;
}

class pibox::mixed {
public:
	class item_t {
	public:
		size_t size;
		char* data;
	};
private:
	size_t size;
	char* ram_begin;
	char* ram_end;
	char* data_begin;
	char* data_end;
	typedef fixed<item_t> item_box_t;
	item_box_t item_box;
	inline static void copy(void* destination, const void* source, size_t size) {
		for (size_t i = 0; i < size; ++i)
			((char*)destination)[i] = ((const char*)source)[i];
	}
public:
	typedef item_box_t::iterator iterator;
	typedef item_box_t::reverse_iterator reverse_iterator;
	inline size_t difference(const item_t*const& minuend, const item_t*const& subtrahent) {
		return item_box.difference(minuend, subtrahent);
	}
	inline size_t difference(iterator const& minuend, iterator const& subtrahent) {
		return item_box.difference(minuend, subtrahent);
	}
	inline size_t difference(reverse_iterator const& minuend, reverse_iterator const& subtrahent) {
		return item_box.difference(minuend, subtrahent);
	}
	inline iterator begin() {
		return item_box.begin();
	}
	inline iterator end() {
		return item_box.end();
	}
	inline reverse_iterator rbegin() {
		return item_box.rbegin();
	}
	inline reverse_iterator rend() {
		return item_box.rend();
	}
	mixed();
	mixed(size_t item_size, size_t data_size);
	~mixed();
	void reserve(size_t item_size, size_t data_size);
	item_t const& make_front(size_t size);
	inline void push_front(item_t item) {
		push_front(item.data, item.size);
	}
	void push_front(const void* data, size_t size);
	item_t shrink_front(size_t size);
	item_t stretch_front(size_t size);
	inline void prepend_front(item_t item) {
		prepend_front(item.data, item.size);
	}
	void prepend_front(const void* data, size_t size);
	item_t& fuse_front();
	inline item_t& fuse_front(size_t size) {
		for (size_t i = 0; i < size; ++i)
			fuse_front();
		return front();
	}
	inline iterator fuse_front(iterator end) {
		for (iterator front = begin(); front != end; front = begin())
			fuse_front();
		return begin();
	}
	item_t const& make_back(size_t size);
	inline void push_back(item_t item) {
		push_back(item.data, item.size);
	}
	void push_back(const void* data, size_t size);
	item_t shrink_back(size_t size);
	item_t stretch_back(size_t size);
	inline void append_back(item_t item) {
		append_back(item.data, item.size);
	}
	void append_back(const void* data, size_t size);
	item_t& fuse_back();
	inline item_t& fuse_back(size_t size) {
		for (size_t i = 0; i < size; ++i)
			fuse_back();
		return back();
	}
	inline reverse_iterator fuse_back(reverse_iterator rend) {
		for (reverse_iterator rback = rbegin(); rback != rend; rback = rbegin())
			fuse_back();
		return rbegin();
	}
	inline item_t& front() {
		return item_box.front();
	}
	inline item_t& front(size_t size) {
		return item_box.front(size);
	}
	inline item_t& back() {
		return item_box.back();
	}
	inline item_t& back(size_t size) {
		return item_box.back(size);
	}
	item_t& fuse(ssize_t size) {
		if (size < 0)
			return fuse_back(-size);
		else
			return fuse_front(size);
	}
	iterator fuse(iterator end) {
		return fuse_front(end);
	}
	reverse_iterator fuse(reverse_iterator rend) {
		return fuse_back(rend);
	}
	inline bool continious(const char*const& front, const char*const& back) {
		return front <= back;
	}
	inline bool continious(iterator front, iterator back) {
		return continious(front->data, back->data);
	}
	inline bool continious(reverse_iterator rback, reverse_iterator rend) {
		return continious(rend->data, rback->data);
	}
	inline bool continious_front(size_t size) {
		item_t* item = &item_box.front();
		return continious(item->data, item_box.increment(item, size)->data);
	}
	inline bool continious_front(size_t size, size_t front) {
		item_t* item = item_box.increment(&item_box.front(), front);
		return continious(item->data, item_box.increment(item, size)->data);
	}
	inline bool continious_back(size_t size) {
		item_t* item = &item_box.back();
		return continious(item->data, item_box.decrement(item, size)->data);
	}
	inline bool continious_back(size_t size, size_t back) {
		item_t* item = item_box.decrement(&item_box.back(), back);
		return continious(item->data, item_box.decrement(item, size)->data);
	}
	inline bool continious(ssize_t size) {
		if (size < 0)
			return continious_back(-size);
		else
			return continious_front(size);
	}
	inline bool continious(ssize_t size, size_t item) {
		if (size < 0)
			return continious_back(-size, item);
		else
			return continious_front(size, item);
	}
	void pop_front();
	void pop_back();
	void clear();
	void free();
	inline size_t item_size() {
		return item_box.size();
	}
	inline size_t item_used() {
		return item_box.used();
	}
	inline size_t item_left() {
		return item_box.left();
	}
	inline size_t data_size() {
		return ram_end-ram_begin;
	}
	inline size_t data_used() {
		return size;
	}
	inline size_t data_left() {
		return data_size()-data_used();
	}
	void data_left(size_t& front, size_t& back);
	inline item_t& operator[](const ssize_t& size) {
		return item_box[size];
	}
};

/*
 *         new_front    front                   back      new_back
 *         |            |                       |         |
 * +-------+~~~~~~~~~~~~+=======+=====+====+====+=========+~~~~----+
 * |                    |                                 |        |
 * ram_begin            data_begin                        data_end ram_end
 *
 *             back new_back     new_front    front
 *             |    |            |            |
 * +======+====+====+~~~~~~~-----+~~~~~~~~~~~~+=========+===+------+
 * |                |                         |                    |
 * ram_begin        data_end                  data_begin           ram_end
 *
 */

#endif
