/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_
#define _PICONF_

#include "piconf.hpp"

#include <stdio.h>
#include <sstream>
#include <string>
#include <vector>
#include <string>
#include <string.h>
#include <stdlib.h>

#define PICONF_WEBSITE "Website: <" << PACKAGE_URL << ">"

piconf::error_t piconf::error;

const char* piconf::SERVICE = "3141";

piconf::error_t::error_t() {
	state = 0;
}

void piconf::error_t::set(state_t what) {
	state |= what;
}

piconf::error_t::state_t piconf::error_t::get(state_t mask) {
	return state & mask;
}

std::string piconf::ts2str(timespec ts) {
	int len = strlen("2012-12-31 12:59:59.123456789")+1;
	char buf[len];
	int ret;
	tm t;
	tzset();
	if (localtime_r(&(ts.tv_sec), &t) == NULL)
		return "Cannot get local time";
	ret = strftime(buf, len, "%F %T", &t);
	if (ret == 0)
		return "Cannot format local time";
	len -= ret-1;
	ret = snprintf(&buf[strlen(buf)], len, ".%09ld", ts.tv_nsec);
	if (ret >= len)
		return "Cannot append nano seconds";
	return buf;
}

std::string piconf::time() {
	if (!VIO::IO::timer)
		return "[No timer specified] ";
	return "["+ts2str(VIO::IO::timer->getLastTime())+"] ";
}

timespec piconf::diff(timespec start, timespec end) {
	timespec temp;
	if ((end.tv_nsec-start.tv_nsec)<0) {
		temp.tv_sec = end.tv_sec-start.tv_sec-1;
		temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
	}
	else {
		temp.tv_sec = end.tv_sec-start.tv_sec;
		temp.tv_nsec = end.tv_nsec-start.tv_nsec;
	}
	return temp;
}

std::string piconf::calculate_size(uint64_t size) {
	static const char* sizes[] = { "EiB", "PiB", "TiB", "GiB", "MiB", "KiB", "B" };
	static const uint64_t exbibytes = 1024ULL * 1024ULL * 1024ULL * 1024ULL * 1024ULL * 1024ULL;
	char result[20];
	uint64_t multiplier = exbibytes;
	for (size_t i = 0; i < sizeof(sizes)/sizeof(*sizes); ++i, multiplier /= 1024)  {   
		if (size < multiplier)
			continue;
		if (size % multiplier == 0)
			sprintf(result, "%" PRIu64 " %s", size/multiplier, sizes[i]);
		else
			sprintf(result, "%.3f %s", (float)size/multiplier, sizes[i]);
		return result;
	}
	return "0 B";
}

std::string piconf::calculate_percentage(size_t value, size_t reference) {
	size_t percentage = 0;
	if (reference)
		percentage = (float)value/reference*100;
	return ((static_cast<std::ostringstream&>(std::ostringstream().flush() << percentage))).str()+"%";
}

void piconf::print_version(name_t name, std::ostream& ansi) {
	ansi << PACKAGE_TARNAME << "-" << PACKAGE_VERSION << std::endl;
	ansi << std::endl;
	ansi << "Copyright (C) 2013  Rouven Spreckels" << "  <" << PACKAGE_BUGREPORT << ">" << std::endl;
	ansi << std::endl;
	ansi << PACKAGE << " is free software: you can redistribute it and/or modify" << std::endl;
	ansi << "it under the terms of the GNU Affero General Public License version 3 as" << std::endl;
	ansi << "published by the Free Software Foundation on 19 November 2007." << std::endl;
	ansi << std::endl;
	ansi << PACKAGE << " is distributed in the hope that it will be useful," << std::endl;
	ansi << "but WITHOUT ANY WARRANTY; without even the implied warranty of" << std::endl;
	ansi << "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << std::endl;
	ansi << "GNU General Public License for more details." << std::endl;
	ansi << std::endl;
	ansi << "You should have received a copy of the GNU Affero General Public License" << std::endl;
	ansi << "along with " << PACKAGE << ".  If not, see <http://www.gnu.org/licenses/>." << std::endl;
	ansi << std::endl;
	ansi << PICONF_WEBSITE << std::endl;
}

void piconf::print_help(name_t name, std::ostream& ansi) {
	ansi << name.s << " (" << name.m << ") - " << name.l << std::endl;
	ansi << std::endl;
	ansi << name.s << " [FILE]" << std::endl;
	ansi << "  With no FILE, or when FILE is -, read standard input." << std::endl;
	ansi << std::endl;
	ansi << name.s << " OPTION" << std::endl;
	ansi << "  -h, --help     Print this help and exit." << std::endl;
	ansi << "  -v, --version  Print version information and exit." << std::endl;
	ansi << std::endl;
	ansi << "For complete documentation, run `man " << name.s << "'." << std::endl;
	ansi << std::endl;
	ansi << PICONF_WEBSITE << std::endl;
}

void piconf::print_info(name_t name, std::ostream& ansi) {
	ansi << "Try `" << name.s << " --help' for more information." << std::endl;
}

std::string piconf::parse_args(name_t name, int argc, char** argv) {
	if (argc == 1)
		return "";
	else
	if (argc == 2) {
		if (!strcmp(argv[1], "--help") or !strcmp(argv[1], "-h"))
			piconf::print_help(name);
		else
		if (!strcmp(argv[1], "--version") or !strcmp(argv[1], "-v"))
			piconf::print_version(name);
		else
			return argv[1];
	}
	else {
		piconf::error.set(error_t::PARSING_ERROR);
		piconf::print_info(name);
	}
	exit(piconf::error.get());
}

#endif
