/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_CONF_TAG_
#define _PICONF_CONF_TAG_

#include "conf_tag.hpp"
#include "../../../ape/ape.hpp"
#include <iostream>
#include <vector>
#include <string>
#include <string.h>

piconf::conf_tag::file_t::file_t() {
	state = 0;
}

piconf::conf_tag::conf_tag(pugi::xml_node node)
: data_tag(node) {
}

bool piconf::conf_tag::instruct(io_tag* io, bool fail, bool exit) {
	return data_tag::instruct(io, fail, exit);
}

void piconf::conf_tag::parse_response(io_tag* io, pugi::xml_node node) {
	data_tag::parse_response(io, node);
	if (file_map.find(io) == file_map.end()) {
		file_box_t file_box;
		file_map.insert(std::make_pair(io, file_box));
		for (rb_box_t::iterator it = rb_box.begin(); it != rb_box.end(); ++it) {
			file_map[io].push_back(file_t());
			std::ostringstream os;
			pugi::xml_node hint = node.first_element_by_path(it->node.c_str());
			if (hint)
				hint.print(os);
			else {
				std::string error = "Node missing in response";
				os << error;
				io_box[io].result.push_back((std::string)"   "+it->node+": "+error);
				piconf::error.set(error_t::XML_TAG_ERROR | error_t::PICONFS_ERROR);
			}
			it->data = os.str();
			file_map[io].back().data = &it->data;
			file_map[io].back().file.assign((it->path+"/"+io->id()+"/"+it->name).c_str());
		}
	}
	parse_result(node, io_box[io].result);
}

void piconf::conf_tag::status(lines_t& lines) {
	lines.push_back(" Merging configuration: done.");
}

void piconf::conf_tag::setup() {
	conf_state = 0;
	ape::merge_twins(node);
	parse_conf(node);
	data_tag::setup();
}

void piconf::conf_tag::parse_conf(pugi::xml_node node) {
	for (pugi::xml_node child = node; child and child.type() == 2; child = child.next_sibling()) {
		std::string path = child.attribute("readback").value();
		if (path.size()) {
			rb_box.push_back(rb_t());
			rb_box.back().path = path;
			rb_box.back().name = child.attribute("name").value();
			rb_box.back().node = child.path();
			rb_box.back().node = rb_box.back().node.substr(((std::string)"/"+PICONF_ROOT+"/pi").size());
		}
		if (child.first_child() and child.first_child().type() == 2)
			parse_conf(child.first_child());
	}
}

void piconf::conf_tag::parse_result(pugi::xml_node node, lines_t& lines, bool file_rb) {
	pugi::xml_attribute rb = node.attribute("readback");
	if (rb) {
		file_rb = strlen(rb.value());
	}
	if (node.first_child() and node.first_child().type() == 2) {
		for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling()) {
			parse_result(child, lines, file_rb);
		}
	}
	else {
		int code = 0;
		std::string value = node.text().get();
		if (!value.size() and node.parent() != node.root()) {
			value = "Unhandled tag";
			code = error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
		}
		code |= node.attribute("error").as_int();
		piconf::error.set(code);
		if (!file_rb or code) {
			lines_t values = explode(value);
			if (values.size()) {
				if (values.size() > 1) {
					lines.push_back((std::string)"   "+node.path().c_str()+":");
					values = indent("    ", values.begin(), values.end());
					lines.insert(lines.end(), values.begin(), values.end());
				}
				else
					lines.push_back((std::string)"   "+node.path().c_str()+": "+values[0]);
			}
		}
	}
}

bool piconf::conf_tag::process() {
	bool undone = false;
	if (!data_tag::process()) {
		for (file_map_t::iterator it = file_map.begin(); it != file_map.end(); ++it) {
			for (file_box_t::iterator ft = it->second.begin(); ft != it->second.end(); ++ft) {
				if (!(ft->state & file_t::DONE)) {
					if (ft->state & file_t::WRITING) {
						if (!ft->file.hasToWrite()) {
							ft->state |= file_t::DONE;
						}
					}
					else {
						try {
							ft->file.remove();
						}
						catch (VIO::IO::Error& error) {
						}
						try {
							ft->file.mkdir();
						}
						catch (VIO::IO::Error& error) {
						}
						try {
							ft->file.open(O_RDWR | O_CREAT);
						}
						catch (VIO::IO::Error& error) {
							io_box[it->first].result.push_back("   /conf: "+(std::string)error.what());
							ft->state |= file_t::DONE;
							piconf::error.set(error_t::FILE_IO_ERROR | error_t::PICONFM_ERROR);
						}
						ft->file.push(ft->data->c_str(), ft->data->size());
						ft->state |= file_t::WRITING;
					}
					if (!(ft->state & file_t::DONE))
						undone = true;
				}
			}
		}
	}	
	else
		undone = true;
	return undone;
}

void piconf::conf_tag::result(io_tag* io, lines_t& lines) {
	data_tag::result(io, lines);
}


#endif
