/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_FILE_TAG_
#define _PICONF_FILE_TAG_

#include "file_tag.hpp"
#include <sstream>

piconf::file_tag::file_tag(pugi::xml_node node)
: data_tag(node) {
	file_data = 0;
	state = 0;
	file.chunk(4096, 4096);
}

void piconf::file_tag::setup() {
	if (node.attribute("readback")) {
		state |= READBACK;
		file_path = node.attribute("readback").value();
	}
	else {
		file.assign(node.text().get());
		file.open();
		file_size = file.size();
		node.append_attribute("size") = ((static_cast<std::ostringstream&>(std::ostringstream().flush() << file_size))).str().c_str();
		state |= FETCH;
	}
	data_tag::setup();
}

bool piconf::file_tag::instruct(io_tag* io, bool fail, bool exit) {
	data_tag::instruct(io, fail, exit);
	if (io_box[io].state & io_t::EXIT)
		return false;
	else
	if (io_box[io].state & PROCESSED_FILE)
		return false;
	else
	if (io_box[io].state & INSTRUCTED_FILE) {
		if (state & READBACK) {
			io_box[io].state |= RECEIVED_FILE;
		}
		else
			return false;
	}
	else
	if (io_box[io].state & io_t::PARSED_RESPONSE) {
		if (io_box[io].state & io_t::VALID_RESPONSE) {
			if (state & READBACK) {
				io->fetch(io_box[io].data_box[1].first, io_box[io].data_box[1].second);
				io_box[io].state |= INSTRUCTED_FILE;
			}
			else {
				if (state & DONE) {
					io->push(file_data, file_size);
					io_box[io].state |= INSTRUCTED_FILE;
				}
			}
		}
		else
			return false;
	}
	return true;
}

void piconf::file_tag::parse_response(io_tag* io, pugi::xml_node node) {
	data_tag::parse_response(io, node);
	std::string error;
	if (node.text())
		error = node.text().get();
	else {
		if (state & READBACK) {
			file_size = node.attribute("size").as_uint();
			if (!file_size) {
				error = "File size of 0 does not make sense";
				piconf::error.set(error_t::XML_TAG_ERROR | error_t::PICONFS_ERROR);
			}
			else
			if (file_size > MAX_FILE_SIZE) {
				error = (std::string)"File size of "+calculate_size(file_size)+" exceeds maximum of "+calculate_size(MAX_XML_SIZE);
				piconf::error.set(error_t::MAXIMUM_ERROR | error_t::PICONFS_ERROR);
			}
			else {
				if (io_box[io].data_box.size() == 1) {
					if (file_path.size())
						file_box.insert(std::make_pair(io, VIO::File()));
					io_box[io].data_box.push_back(std::make_pair(new char[file_size], file_size));
				}
				io_box[io].state |= io_t::VALID_RESPONSE;
			}
		}
		else
			io_box[io].state |= io_t::VALID_RESPONSE;
	}
	if (error.size())
		io_box[io].result.push_back((std::string)"   "+node.path().c_str()+": "+error);
}

bool piconf::file_tag::complete() {
	return state & DONE;
}

bool piconf::file_tag::process() {
	bool undone = false;
	if (state & READBACK) {
		for (io_box_t::iterator it = io_box.begin(); it != io_box.end(); ++it) {
			if (!(it->second.state & PROCESSED_FILE)) {
				if (it->second.state & io_t::PARSED_RESPONSE and !(it->second.state & io_t::VALID_RESPONSE))
					it->second.state |= PROCESSED_FILE;
				if (it->second.state & RECEIVED_FILE) {
					if (file_path.size()) {
						if (it->second.state & WRITING_FILE) {
							if (!file_box[it->first].hasToWrite()) {
								it->second.state |= PROCESSED_FILE;
							}
						}
						else {
							file_box[it->first].assign((file_path+"/"+it->first->id()+"/"+node.attribute("name").value()).c_str());
							try {
								file_box[it->first].remove();
							}
							catch (VIO::IO::Error& error) {
							}
							try {
								file_box[it->first].mkdir();
							}
							catch (VIO::IO::Error& error) {
							}
							try {
								file_box[it->first].open(O_RDWR | O_CREAT);
								file_box[it->first].push(io_box[it->first].data_box[1].first, io_box[it->first].data_box[1].second);
								it->second.state |= WRITING_FILE;
							}
							catch (VIO::IO::Error& error) {
								piconf::error.set(error_t::FILE_IO_ERROR | error_t::PICONFM_ERROR);
								it->second.result.push_back("   /file: "+(std::string)error.what());
								it->second.state |= PROCESSED_FILE;
							}
						}
					}
					else {
						it->second.data_box[1].first[file_size-1] = '\0';
						lines_t lines = explode(it->second.data_box[1].first);
						if (lines.size()) {
							it->second.result.push_back("   /file: "+lines[0]);
							if (lines.size() > 1) {
								lines = indent("          ", lines.begin()+1, lines.end());
								it->second.result.insert(it->second.result.end(), lines.begin(), lines.end());
							}
						}
						it->second.state |= PROCESSED_FILE;
					}
				}
				if (!(it->second.state & PROCESSED_FILE))
					undone = true;
			}
		}
	}
	else {
		undone = true;
		if (state & FETCH) {
			state &= ~FETCH;
			try {
				file_data = new char[file_size];
			}
			catch (std::bad_alloc& error) {
				throw (std::string)"No RAM left to reserve file of "+calculate_size(file_size);
			}
			file.fetch(file_data, file_size);
		}
		if (!(state & DONE)) {
			if (!file.hasToRead()) {
				state |= DONE;
			}
		}
		if (state & DONE)
			undone = false;
	}
	if (data_tag::process())
		undone = true;
	return undone;
}

void piconf::file_tag::status(lines_t& lines) {
	if (!(state & READBACK)) {
		size_t load = file_size-file.hasToRead();
		std::string load_str = calculate_size(load);
		std::string percentage_str = calculate_percentage(load, file_size);
		std::string state = load == file_size ? ", done." : "";
		lines.push_back(" Loading file: "+percentage_str+", "+load_str+state);
		lines.push_back((std::string)"  "+node.text().get());
	}
}

piconf::file_tag::~file_tag() {
	delete[] file_data;
}

#endif
