/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_TCP_TAG_
#define _PICONF_TCP_TAG_

#include "tcp_tag.hpp"
#include "../piconfm.hpp"
#include "../../../vio/TCP.hpp"
#include <string.h>

const timespec piconf::tcp_tag::delay = {3, 0};

piconf::tcp_tag::tcp_tag(pugi::xml_node node)
: io_tag(node) {
}

void piconf::tcp_tag::setup() {
	pugi::xml_node host = node.child("host");
	pugi::xml_node service = node.child("service");
	if (!service) {
		node.append_child("service").text() = SERVICE;
		service = node.child("service");
	}
	if (host and service) {
		this->host = host.text().get();
		this->service = service.text().get();
	}
	else {
		piconf::error.set(error_t::PARSING_ERROR | error_t::PICONFM_ERROR);
		throw (std::string)"Incomplete TCP tag";
	}
	timeout.tv_sec = node.child("timeout").text().as_int();
	if (!timeout.tv_sec)
		timeout.tv_sec = piconf::TIMEOUT;
	timeout.tv_nsec = 0;
	reconnect_delay.tv_sec = node.child("reconnect_delay").text().as_int();
	if (!reconnect_delay.tv_sec)
		reconnect_delay.tv_sec = piconf::RECONNECT_DELAY;
	reconnect_delay.tv_nsec = 0;
	reconnect_delay_begin.tv_sec = 0;
	reconnect_delay_begin.tv_nsec = 0;
	pugi::xml_node tries = node.child("reconnect_tries");
	if (tries)
		reconnect_tries = tries.text().as_uint();
	else
		reconnect_tries = 3;
	reconnect_limit = true;
	reconnect_count = 0;
	if (!reconnect_tries and strcmp(tries.text().get(), "i") >= 0)
		reconnect_limit = false;
	code = 0;
	tcp.owner(this);
	io = &tcp;
	tcp_state = 0;
}

void piconf::tcp_tag::handle(int code, std::string what) {
	tcp_state = 0;
	tcp.clear();
	if (++reconnect_count > reconnect_tries and reconnect_limit) {
		reconnect_count = reconnect_tries;
		switch (code) {
			case VIO::TCP::UNAVAILABLE: code = error_t::PICONFS_ERROR;
			case VIO::TCP::GETADDRINFO: code = error_t::PICONFM_ERROR;
			case VIO::TCP::GETPEERNAME: code = error_t::PICONFM_ERROR;
			case VIO::TCP::GETNAMEINFO: code = error_t::PICONFM_ERROR;
			case VIO::TCP::READ: code = error_t::PICONFS_ERROR;
			case VIO::TCP::WRITE: code = error_t::PICONFS_ERROR;
			case 0: code = error_t::PICONFS_ERROR;
			default: code = error_t::PICONFM_ERROR;
		}
		piconf::error.set(error_t::PEER_IO_ERROR | code);
		tcp_state |= FAIL;
		instruct(true, true);
	}
	else
		tcp.wait(reconnect_delay_begin);
	this->what = what;
	if (reconnect_tries or !reconnect_limit) {
		std::string total = reconnect_limit ? to_string(reconnect_tries) : "-";
		this->what += ", retry ("+to_string(reconnect_count)+"/"+total+")";
	}
}

void piconf::tcp_tag::schedule(const timespec* interval, const timespec* realtime, const timespec* accuracy) {
}

bool piconf::tcp_tag::process() {
	if (tcp_state & DONE or tcp_state & FAIL)
		return false;
	if (tcp_state and tcp.isConnected(timeout)) {
		if (tcp_state & CONNECTING) {
			tcp_state &= ~CONNECTING;
			tcp_state |= CONNECTED;
			reconnect_count = 0;
			instruct(true);
			what = "";
		}
		else {
			if (!io_tag::process()) {
				try {
					tcp.close();
				}
				catch (VIO::IO::Error& error) {
					what = error.what();
				}
				tcp.clear();
				tcp_state |= DONE;
			}
		}
	}
	else {
		if (tcp_state & CONNECTED) {
			tcp_state &= ~CONNECTED;
			handle(0, "Lost connection");
		}
		if (!(tcp_state & CONNECTING)) {
			if (!tcp.hasToWait(reconnect_delay_begin, reconnect_delay)) {
				tcp.clear();
				tcp_state |= CONNECTING;
				tcp.connect(host, service);
			}
		}
	}
	return true;
}

void piconf::tcp_tag::status(lines_t& lines) {
	std::string data_done = to_string(data+(data < data_box.size() ? 1 : 0));
	std::string data_size = to_string(data_box.size());
	lines.push_back((std::string)" Transfering tag ("+data_done+"/"+data_size+") with "+host+" at "+service+":");
	lines.push_back("");
	{
		std::string percentage = calculate_percentage(o_progress.tag.done, o_progress.tag.size);
		std::string size = calculate_size(o_progress.tag.done);
		lines.back().append((std::string)"  "+percentage+" ("+size+") sent");
	}
	{
		std::string percentage = calculate_percentage(i_progress.tag.done, i_progress.tag.size);
		std::string size = calculate_size(i_progress.tag.done);
		lines.back().append((std::string)", "+percentage+" ("+size+") received");
	}
	if (tcp_state & DONE) {
		lines.back().append((std::string)", done.");
	}
	else
	if (tcp_state & FAIL) {
		lines.back().append((std::string)", aborted.");
	}
	else
	if (tcp.isConnected()) {
		lines.back().append((std::string)", connected.");
	}
	else {
		lines.back().append((std::string)", connecting...");
	}
	if (what.size()) {
		lines.push_back((std::string)"  WARNING: "+what);
	}
	io_tag::status(lines);
}

std::string piconf::tcp_tag::id() {
	if (host.find(":") != std::string::npos)
		return (std::string)"["+host+"]:"+service;
	else
		return host+":"+service;
}

piconf::tcp_tag::~tcp_tag() {
}

#endif
