/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_S_
#define _PICONF_S_

#include "piconfs.hpp"

#include <string>
#include <vector>
#include <stdio.h>
#include <string.h>

piconf::name_t piconf::name = {
	PICONF_NAME_S "s",
	PICONF_NAME_M " slave",
	PICONF_NAME_L " (Slave)"
};

int piconf::shell(line_t& output, line_t input, chunk_t size) {
	FILE* stream;
	if (!(stream = ::popen((input+" 2>&1").c_str(), "r")))
		throw std::string("Shell failed to run: "+input);
	char chunk[size];
	while (NULL != fgets(chunk, sizeof chunk, stream))
		output += chunk;
	output = output.substr(0, output.size()-1);
	return pclose(stream);
}

void piconf::axi_gpio(pugi::xml_node node, axi_map_box_t& zdx_box, axi_buf_box_t& buf_box, std::ofstream& log) {
	ape::inherit_attributes(node);
	static axi_box_t axi_box;
	axi_ptr_box_t hsi_box;
	axi_ptr_box_t hso_box;
	axi_cat_box_t zdi_box;
	axi_cat_box_t zdo_box;
	class AXI {
	public:
		static VIO::AXI& setup(VIO::AXI& back, pugi::xml_node node) {
			pugi::xml_attribute device = node.attribute("device");
			pugi::xml_attribute hsx_delay = node.attribute("hsx_delay");
			pugi::xml_attribute zdx_delay = node.attribute("zdx_delay");
			if (device)
				back.device = device.value();
			if (hsx_delay)
				back.hsx_delay = hsx_delay.as_uint();
			if (zdx_delay)
				back.zdx_delay = zdx_delay.as_uint();
			return back;
		}
	};
	for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling()) {
		if (!strcmp(child.name(), "hsi")) {
			if (hsi_box.size())
				throw (std::string)"Only one hsi is allowed. You should append further axi instances instead.";
			log << time() << "Generating hsi[" << (hsi_box.size()+1) << "]:" << std::endl;
			for (pugi::xml_node axi = child.first_child(); axi and axi.type() == 2; axi = axi.next_sibling()) {
				pugi::xml_attribute address = axi.attribute("address");
				if (!address)
					throw (std::string)"No base address specified for axi["+to_string(hsi_box.size()+1)+"] of hsi[1].";
				axi_box.push_back(VIO::AXI());
				AXI::setup(axi_box.back(), axi);
				axi_box.back().open(to_number<off_t>(address.value()));
				log << time() << " Instantiated axi[" << (hsi_box.size()+1) << "] with address " << address.value() << " for hsi[1]." << std::endl;
				hsi_box.push_back(&axi_box.back());
			}
		}
		else
		if (!strcmp(child.name(), "hso")) {
			if (hso_box.size())
				throw (std::string)"Only one hso is allowed. You should append further axi instances instead.";
			log << time() << "Generating hso[" << (hso_box.size()+1) << "]:" << std::endl;
			for (pugi::xml_node axi = child.first_child(); axi and axi.type() == 2; axi = axi.next_sibling()) {
				pugi::xml_attribute address = axi.attribute("address");
				if (!address)
					throw (std::string)"No base address specified for axi["+to_string(hso_box.size()+1)+"] of hso[1].";
				axi_box.push_back(VIO::AXI());
				AXI::setup(axi_box.back(), axi);
				axi_box.back().open(to_number<off_t>(address.value()));
				log << time() << " Instantiated axi[" << (hso_box.size()+1) << "] with address " << address.value() << " for hso[1]." << std::endl;
				hso_box.push_back(&axi_box.back());
			}
		}
		else
		if (!strcmp(child.name(), "zdi")) {
			axi_ptr_box_t cat_box;
			log << time() << "Generating zdi[" << (zdi_box.size()+1) << "]:" << std::endl;
			size_t bit = child.attribute("bit").as_uint();
			size_t pos = bit/sizeof (uint32_t);
			size_t min = hsi_box.size();
			if (hso_box.size() < min)
				min = hso_box.size();
			if (!(pos < min))
				throw (std::string)"hsx["+to_string(pos+1)+"] does not yet exist.";
			bit -= pos*sizeof (uint32_t);
			pugi::xml_attribute purpose = child.attribute("purpose");
			if (!purpose or !strlen(purpose.value()))
				throw (std::string)"No purpose for zdi specified";
			for (pugi::xml_node axi = child.first_child(); axi and axi.type() == 2; axi = axi.next_sibling()) {
				pugi::xml_attribute address = axi.attribute("address");
				if (!address)
					throw (std::string)"No base address specified for axi["+to_string(cat_box.size()+1)+"] of zdi["+to_string(zdi_box.size()+1)+"].";
				log << time() << " Instantiating axi[" << (cat_box.size()+1) << "] with address " << address.value() << " ";
				if (cat_box.size()) {
					axi_box.push_back(VIO::AXI());
					AXI::setup(axi_box.back(), axi);
					axi_box.back().open(to_number<off_t>(address.value()));
					cat_box.back()->next(&axi_box.back());
					log << "as nxt[" << (cat_box.size()+1) << "] for zdi[" << (zdi_box.size()+1) << "]." << std::endl;
				}
				else {
					axi_box.push_back(VIO::AXI());
					AXI::setup(axi_box.back(), axi);
					axi_box.back().open(to_number<off_t>(address.value()), hsi_box[pos], hso_box[pos], bit, bit);
					zdx_box.insert(std::make_pair<std::string, VIO::AXI*>(purpose.value(), &axi_box.back()));
					log << "and hsx[1][" << (pos+1) << "]:" << bit << " for zdi[" << (zdi_box.size()+1) << "]." << std::endl;
				}
				cat_box.push_back(&axi_box.back());
			}
			zdi_box.push_back(cat_box);
			buf_box.insert(std::make_pair<std::string, size_t>(purpose.value(), cat_box.size()*sizeof (uint32_t)));
		}
		else
		if (!strcmp(child.name(), "zdo")) {
			axi_ptr_box_t cat_box;
			log << time() << "Generating zdo[" << (zdo_box.size()+1) << "]:" << std::endl;
			size_t bit = child.attribute("bit").as_uint();
			size_t pos = bit/sizeof (uint32_t);
			size_t min = hsi_box.size();
			if (hso_box.size() < min)
				min = hso_box.size();
			if (!(pos < min))
				throw (std::string)"hsx["+to_string(pos+1)+"] does not yet exist.";
			bit -= pos*sizeof (uint32_t);
			pugi::xml_attribute purpose = child.attribute("purpose");
			if (!purpose or !strlen(purpose.value()))
				throw (std::string)"No purpose for zdo specified";
			for (pugi::xml_node axi = child.first_child(); axi and axi.type() == 2; axi = axi.next_sibling()) {
				pugi::xml_attribute address = axi.attribute("address");
				if (!address)
					throw (std::string)"No base address specified for axi["+to_string(cat_box.size()+1)+"] of zdo["+to_string(zdo_box.size()+1)+"].";
				log << time() << " Instantiating axi[" << (cat_box.size()+1) << "] with address " << address.value() << " ";
				if (cat_box.size()) {
					axi_box.push_back(VIO::AXI());
					AXI::setup(axi_box.back(), axi);
					axi_box.back().open(to_number<off_t>(address.value()));
					cat_box.back()->next(&axi_box.back());
					log << "as nxt[" << (cat_box.size()+1) << "] for zdo[" << (zdo_box.size()+1) << "]." << std::endl;
				}
				else {
					axi_box.push_back(VIO::AXI());
					AXI::setup(axi_box.back(), axi);
					axi_box.back().open(to_number<off_t>(address.value()), hsi_box[pos], hso_box[pos], bit, bit);
					zdx_box.insert(std::make_pair<std::string, VIO::AXI*>(purpose.value(), &axi_box.back()));
					log << "and hsx[1][" << (pos+1) << "]:" << bit << " for zdo[" << (zdo_box.size()+1) << "]." << std::endl;
				}
				cat_box.push_back(&axi_box.back());
			}
			zdo_box.push_back(cat_box);
			buf_box.insert(std::make_pair<std::string, size_t>(purpose.value(), cat_box.size()*sizeof (uint32_t)));
		}
		else
			throw (std::string)"Invalid node '"+child.name()+"' inside io node";
	}
}

#endif
