/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_CONF_TAG_
#define _PICONF_CONF_TAG_

#include "conf_tag.hpp"
#include <string.h>

piconf::axi_map_box_t piconf::conf_tag::zdx_box;

pugi::xml_node piconf::conf_tag::tpl;

char* piconf::conf_tag::data = 0;

void piconf::conf_tag::construct(std::ofstream& log, pugi::xml_node tpl) {
	conf_tag::tpl = tpl;
	try {
		pugi::xml_node io = tpl.child("io");
		if (io) {
			axi_buf_box_t buf_box;
			if (strcmp(io.attribute("protocol").value(), "AXI_GPIO"))
				throw (std::string)"Unknown io protocol '"+io.attribute("protocol").value()+"'";
			axi_gpio(io, zdx_box, buf_box, log);
			for (axi_buf_box_t::iterator it = buf_box.begin(); it != buf_box.end(); ++it)
				if (it->second > MAX_AXI_GPIO_SIZE)
					throw (std::string)"AXI size of "+calculate_size(it->second)+" exceeds maximum of "+calculate_size(MAX_AXI_GPIO_SIZE);
			if (zdx_box.find("i") == zdx_box.end())
				throw (std::string)"No zdi with purpose 'i' found";
			if (zdx_box.find("o") == zdx_box.end())
				throw (std::string)"No zdo with purpose 'o' found";
			if (!buf_box["i"])
				throw (std::string)"AXI buffer has size of zero";
			if (buf_box["i"] != buf_box["o"])
				throw (std::string)"AXI buffer sizes differ";
			size_t bit = highest_bit(tpl.child("data"));
			log << time() << "Determined highest bit of " << bit << std::endl;
			size_t axi_size = 8*buf_box["i"];
			if (bit >= axi_size)
				throw (std::string)"Highest bit of "+to_string(bit)+" exceeds size of "+to_string(axi_size)+" b";
			data = new char[buf_box["i"]];
			memset(data, 0, buf_box["i"]);
			log << time() << "Allocated data buffer of " << buf_box["i"] << " B for " << axi_size << " b" << std::endl;
		}
	}
	catch (VIO::AXI::Error& error) {
		throw (std::string)error.what();
	}
}

piconf::conf_tag::conf_tag(std::ofstream& log, pugi::xml_node tpl, pugi::xml_node tag, VIO::IO* io)
: data_tag(log, tag, io) {
	try {
		write(data);
		if (data) {
			//zdx_box["o"]->write((const uint32_t*)data);
			//zdx_box["i"]->read((uint32_t*)data);
		}
		read(data);
	}
	catch (VIO::AXI::Error& error) {
		throw (std::string)error.what();
	}
	if (!io)
		return;
	respond();
	log << time() << "Parsed conf_tag" << std::endl;
}

size_t piconf::conf_tag::highest_bit(pugi::xml_node node) {
	size_t bit = node.attribute("bit").as_uint();
	for (pugi::xml_node child = node.first_child(); child and child.type() == 2; child = child.next_sibling()) {
		size_t child_bit = highest_bit(child);
		if (child_bit > bit)
			bit = child_bit;
	}
	return bit;
}

void piconf::conf_tag::write(void* data, pugi::xml_node tpl, pugi::xml_node tag) {
	if (tag.attribute("readback"))
		return;
	for (pugi::xml_node tag_child = tag.first_child(); tag_child and tag_child.type() == 2;) {
		if (tag_child.attribute("readback")) {
			tag_child = tag_child.next_sibling();
			continue;
		}
		pugi::xml_node tpl_child = tpl.child(tag_child.name());
		if (tpl_child) {
			if (tag_child.first_child() and tag_child.first_child().type() == 2) {
				write(data, tpl_child, tag_child);
				if (!(tag_child.first_child() and tag_child.first_child().type() == 2)) {
					pugi::xml_node temp = tag_child;
					tag_child = tag_child.next_sibling();
					tag.remove_child(temp);
					continue;
				}
			}
			else
			if (tpl_child.first_child() and tpl_child.first_child().type() == 2) {
				pugi::xml_node temp = tag_child;
				tag_child = tag_child.next_sibling();
				tag.remove_child(temp);
				continue;
			}
			else {
				try {
					std::string error;
					if (!strcmp(tpl_child.attribute("purpose").value(), "action")) {
						uint32_t number = to_number<uint32_t>(tag_child.text().get());
						error = in_range(
							number,
							tpl_child.attribute("min").as_uint(),
							tpl_child.attribute("max").as_uint()
						);
						if (data and number) {
							if (!error.size())
								error = get_uint(
									data,
									number,
									tpl_child.attribute("bit").as_uint(),
									tpl_child.attribute("bits").as_uint(),
									tpl_child.attribute("min").as_uint(),
									tpl_child.attribute("max").as_uint()
								);
							if (!error.size())
								error = set_uint(
									data,
									(uint8_t)(number ? 0 : 1),
									tpl_child.attribute("bit").as_uint(),
									tpl_child.attribute("bits").as_uint(),
									tpl_child.attribute("min").as_uint(),
									tpl_child.attribute("max").as_uint()
								);
						}
					}
					else
					if (data and !strcmp(tpl_child.attribute("purpose").value(), "config")) {
						uint32_t number = to_number<uint32_t>(tag_child.text().get());
						error = set_uint(
							data,
							number,
							tpl_child.attribute("bit").as_uint(),
							tpl_child.attribute("bits").as_uint(),
							tpl_child.attribute("min").as_uint(),
							tpl_child.attribute("max").as_uint()
						);

					}
					else {
						error = "Readback only tag";
					}
					if (error.size()) {
						tag_child.text().set(error.c_str());
						if (!tag_child.attribute("error"))
							tag_child.append_attribute("error");
						int code = error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
						tag_child.attribute("error").set_value(to_string(code).c_str());
						piconf::error.set(code);
					}
					else {
						pugi::xml_node temp = tag_child;
						tag_child = tag_child.next_sibling();
						tag.remove_child(temp);
						continue;
					}
				}
				catch (std::string& error) {
					tag_child.text().set(error.c_str());
					if (!tag_child.attribute("error"))
						tag_child.append_attribute("error");
					int code = error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
					tag_child.attribute("error").set_value(to_string(code).c_str());
					piconf::error.set(code);
				}
			}
		}
		else {
			tag_child.text().set("Tag unknown");
			if (!tag_child.attribute("error"))
				tag_child.append_attribute("error");
			int code = error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
			tag_child.attribute("error").set_value(to_string(code).c_str());
			piconf::error.set(code);
			for (pugi::xml_node tag_grandchild = tag_child.first_child(); tag_grandchild and tag_grandchild.type() == 2; tag_grandchild = tag_grandchild.next_sibling()) {
				tag_child.remove_child(tag_grandchild);
			}
		}
		tag_child = tag_child.next_sibling();
	}
}

void piconf::conf_tag::read(const void* data, pugi::xml_node tpl, pugi::xml_node tag, bool readback) {
	if (tag.attribute("readback"))
		readback = true;
	if (readback) {
		for (pugi::xml_node tag_child = tag.first_child(); tag_child and tag_child.type() == 2; tag_child = tag_child.next_sibling()) {
			if (!tpl.child(tag_child.name())) {
				tag_child.text().set("Tag unknown");
				if (!tag_child.attribute("error"))
					tag_child.append_attribute("error");
				int code = error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
				tag_child.attribute("error").set_value(to_string(code).c_str());
				piconf::error.set(code);
				for (pugi::xml_node tag_grandchild = tag_child.first_child(); tag_grandchild and tag_grandchild.type() == 2; tag_grandchild = tag_grandchild.next_sibling()) {
					tag_child.remove_child(tag_grandchild);
				}
			}
		}
		if (tpl.first_child() and tpl.first_child().type() == 2) {
			for (pugi::xml_node tpl_child = tpl.first_child(); tpl_child and tpl_child.type() == 2; tpl_child = tpl_child.next_sibling()) {
				if (
					!strcmp(tpl_child.attribute("purpose").value(), "action") or
					!strcmp(tpl_child.attribute("purpose").value(), "status") or
					!strcmp(tpl_child.attribute("purpose").value(), "config")
				) {
					pugi::xml_node tag_child = tag.child(tpl_child.name());
					if (!tag_child) {
						tag_child = tag.append_child();
						tag_child.set_name(tpl_child.name());
					}
					read(data, tpl_child, tag_child, true);
				}
			}
		}
		else {
			if (
				!strcmp(tpl.attribute("purpose").value(), "status") or
				!strcmp(tpl.attribute("purpose").value(), "config")
			) {
				uint32_t number;
				if (data)
					get_uint(
						data,
						number,
						tpl.attribute("bit").as_uint(),
						tpl.attribute("bits").as_uint(),
						tpl.attribute("min").as_uint(),
						tpl.attribute("max").as_uint()
					);
				tag.text().set(to_string(number).c_str());
			}
			else {
				tag.text().set(tpl.attribute("min").value());
			}
		}
	}
	else {
		for (pugi::xml_node tag_child = tag.first_child(); tag_child and tag_child.type() == 2; tag_child = tag_child.next_sibling()) {
			pugi::xml_node tpl_child = tpl.child(tag_child.name());
			if (tpl_child) {
				read(data, tpl_child, tag_child);
			}
			else {
				tag_child.text().set("Tag unknown");
				if (!tag_child.attribute("error"))
					tag_child.append_attribute("error");
				int code = error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
				tag_child.attribute("error").set_value(to_string(code).c_str());
				piconf::error.set(code);
				for (pugi::xml_node tag_grandchild = tag_child.first_child(); tag_grandchild and tag_grandchild.type() == 2; tag_grandchild = tag_grandchild.next_sibling()) {
					tag_child.remove_child(tag_grandchild);
				}
			}
		}
	}
}

void piconf::conf_tag::write(void* data) {
	write(data, tpl.child("data"), tag);
}

void piconf::conf_tag::read(const void* data) {
	read(data, tpl.child("data"), tag);
}

void piconf::conf_tag::destruct() {
	delete[] data;
	data = 0;
}

#endif
