/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_DAQ_TAG_
#define _PICONF_DAQ_TAG_

#include "../piconfs.hpp"
#include "daq_tag.hpp"
#include <string>
#include <string.h>

#define PICONF_DAQ_AXI_LOOP 1

//daq-query (daq_tag() constructor)
//daq-value without blocking if nothing to read?

piconf::axi_map_box_t piconf::daq_tag::zdx_box;

pugi::xml_node piconf::daq_tag::tpl;

timespec piconf::daq_tag::activate = {piconf::ACTIVATE, 0};

piconf::daq_tag::daq_state_t piconf::daq_tag::daq_state = 0;

char* piconf::daq_tag::data = 0;

size_t piconf::daq_tag::item_sync = 0;
size_t piconf::daq_tag::item_size = 0;
size_t piconf::daq_tag::ring_size = 0;

size_t piconf::daq_tag::r_item_done = 0;
size_t piconf::daq_tag::i_item_done = 0;
size_t piconf::daq_tag::o_item_done = 0;

size_t piconf::daq_tag::data_size = 0;

pibox::mixed piconf::daq_tag::ring;

piconf::data_tag::hooked_t piconf::daq_tag::hooked = { daq_tag::hook };

void piconf::daq_tag::construct(std::ofstream& log, pugi::xml_node tpl, hooked_t*& hooked, timespec activate) {
	daq_state = 0;
	hooked = &daq_tag::hooked;
	daq_tag::activate = activate;
	daq_tag::tpl = tpl;
	try {
		pugi::xml_node io = tpl.child("io");
		if (io) {
			axi_buf_box_t buf_box;
			if (strcmp(io.attribute("protocol").value(), "AXI_GPIO"))
				throw (std::string)"Unknown io protocol '"+io.attribute("protocol").value()+"'";
			axi_gpio(io, zdx_box, buf_box, log);
			for (axi_buf_box_t::iterator it = buf_box.begin(); it != buf_box.end(); ++it)
				if (it->second > MAX_AXI_GPIO_SIZE)
					throw (std::string)"AXI size of "+calculate_size(it->second)+" exceeds maximum of "+calculate_size(MAX_AXI_GPIO_SIZE);
			if (zdx_box.find("daq") == zdx_box.end())
				throw (std::string)"No zdi with purpose 'daq' found";
			data_size = buf_box["daq"];
			if (!data_size)
				throw (std::string)"AXI buffer has size of zero";
			try {
				data = new char[data_size];
			}
			catch (std::bad_alloc& error) {
				throw (std::string)"No RAM left to reserve AXI buffer of "+calculate_size(data_size);
			}
			log << time() << "Allocated AXI buffer of " << calculate_size(data_size) << std::endl;
		}
	}
	catch (VIO::AXI::Error& error) {
		throw (std::string)error.what();
	}
	ring_size = tpl.attribute("ring_size").as_uint();
	if (!ring_size)
		throw (std::string)"Size of ring buffer is zero";
	if (ring_size > MAX_RING_SIZE)
		throw (std::string)"Size of ring buffer "+calculate_size(ring_size)+" exceeds maximum of "+calculate_size(MAX_RING_SIZE);
	item_size = tpl.attribute("item_size").as_uint();
	if (!item_size)
		throw (std::string)"Size of item buffer is zero";
	if (item_size > MAX_ITEM_SIZE)
		throw (std::string)"Size of item buffer "+to_string(item_size)+" exceeds maximum of "+calculate_size(MAX_ITEM_SIZE);
	item_sync = tpl.attribute("item_sync").as_uint();
	if (!item_sync)
		item_sync = item_size;
	else
	if (item_sync > item_size)
		throw (std::string)"Synchronisation item of "+to_string(item_sync)+" exceeds item size of "+to_string(item_size);
	ring.reserve(item_size, ring_size);
	log << time() << "Allocated ring buffer of " << calculate_size(ring_size) << " with item buffer of " << item_size << std::endl;
}

piconf::daq_tag::daq_tag(std::ofstream& log, pugi::xml_node tpl, pugi::xml_node tag, VIO::IO* io)
: data_tag(log, tag, io) {
	int code = 0;
	daq_state &= ~(INVALID | HOOK_UP);
	if (tag.child("file"))
		daq_state |= HOOK_UP;
	for (pugi::xml_node child = tag.first_child(); child and child.type() == 2; child = child.next_sibling())
		tag.remove_child(child);
	if (daq_state & HOOK_UP and hooked.ios.size()) {
		daq_state &= ~HOOK_UP;
		tag.text() = "Already acquiring data";
		code |= error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
		daq_state |= INVALID;
	}
	if (!(daq_state & INVALID)) {
		if (daq_state & HOOK_UP) {
			o_item_done = tag.attribute("o_item_done").as_uint();
			if (daq_state & STALLED) {
			}
			else {
				if (o_item_done > i_item_done) {
					tag.text() = ((std::string)"o_item_done ("+to_string(o_item_done)+") > i_item_done ("+to_string(i_item_done)+")").c_str();
					code |= error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
					daq_state |= INVALID;
				}
				else {
					daq_state |= SYNCING;
					i_item_done -= o_item_done;
					o_item_done = 0;
				}
			}
		}
	}
	if (!(daq_state & INVALID)) {
		//pugi::xml_attribute readback = tag.attribute("readback");
		//if (readback) {
			tag.append_child("item_size").text() = (calculate_percentage(ring.item_size(), MAX_ITEM_SIZE)+" ("+calculate_size(ring.item_size())+")").c_str();
			tag.child("item_size").append_attribute("value") = ring.item_size();
			tag.append_child("item_used").text() = (calculate_percentage(ring.item_used(), ring.item_size())+" ("+calculate_size(ring.item_used())+")").c_str();
			tag.child("item_used").append_attribute("value") = ring.item_used();
			tag.append_child("item_left").text() = (calculate_percentage(ring.item_left(), ring.item_size())+" ("+calculate_size(ring.item_left())+")").c_str();
			tag.child("item_left").append_attribute("value") = ring.item_left();
			tag.append_child("data_size").text() = (calculate_percentage(ring.data_size(), MAX_RING_SIZE)+" ("+calculate_size(ring.data_size())+")").c_str();
			tag.child("data_size").append_attribute("value") = ring.data_size();
			tag.append_child("data_used").text() = (calculate_percentage(ring.data_used(), ring.data_size())+" ("+calculate_size(ring.data_used())+")").c_str();
			tag.child("data_used").append_attribute("value") = ring.data_used();
			tag.append_child("data_left").text() = (calculate_percentage(ring.data_left(), ring.data_size())+" ("+calculate_size(ring.data_left())+")").c_str();
			tag.child("data_left").append_attribute("value") = ring.data_left();
			tag.append_child("acquire").text() = daq_state & ACQUIRE ? "1" : "0";
			tag.append_child("readout").text() = daq_state & READOUT ? "1" : "0";
			tag.append_child("emptied").text() = daq_state & EMPTIED ? "1" : "0";
			tag.append_child("overrun").text() = daq_state & OVERRUN ? "1" : "0";
		//}
		if (tag.attribute("acquire").as_bool())
			daq_state |= ACQUIRE;
		else
			daq_state &= ~ACQUIRE;
		if (tag.attribute("readout").as_bool())
			daq_state |= READOUT;
		else
			daq_state &= ~READOUT;
		if (tag.attribute("discard").as_bool()) {
			ring.clear();
			daq_state |= EMPTIED;
		}
	}
	else {
		if (!tag.attribute("error"))
			tag.append_attribute("error");
		tag.attribute("error").set_value(to_string(code).c_str());
	}
	respond();
	log << time() << "Parsed daq_tag" << std::endl;
}

bool piconf::daq_tag::hook_me_up() {
	if (daq_state & HOOK_UP) {
		hooked.ios.push_back(io);
		return true;
	}
	return false;
}

void piconf::daq_tag::hook() {
	if (daq_state & UPGRADE) {
		daq_state &= ~UPGRADE;
		queued.insert(queued.begin(), hooked.ios.begin(), hooked.ios.end());
	}
	VIO::IO* io = 0;
	if (hooked.ios.size()) {
		io = hooked.ios.front();
		if (!io)
			hooked.ios.pop_front();
	}
	if (io) {
		if (io->activate(activate)) {
			static uint32_t data[4];
			data[0] = htonl(0);
			data[1] = htonl(daq_state);
			data[2] = htonl(ring.item_used());
			data[3] = htonl(ring.data_used());
			if (!io->hasToWrite())
				io->push(&data, sizeof data);
		}
	}
	if (daq_state & OVERRUN) {
		return;//????
	}
	if (daq_state & ACQUIRE) {//tell it also the pl, yes but also possible via cc->lda!
		if (!data)
			return;
		zdx_box["daq"]->read((uint32_t*)data);
		size_t what = ((uint32_t*)data)[0];
		if (what & (1 << 0)) {
			daq_state |= OVERRUN;
			return;
		}
		static size_t size;
		static pibox::mixed::iterator r_item, i_item = ring.begin(), o_item = i_item;
		static size_t body;
		const size_t head = 4*sizeof (size_t);
		if (daq_state & EMPTIED) {
			daq_state &= ~(EMPTIED | STALLED);
			o_item_done = i_item_done = 0;
			o_item = i_item = ring.begin();
		}
		if (!(daq_state & READING)) {
			size = ntohl(((uint32_t*)data)[1]);
			if (size) {
				bool make = false;
				if (!(o_item == i_item and daq_state & WRITING))
					try {
						ring.stretch_back(size+head);
					}
					catch (std::string& error) {
						make = true;
					}
				else
					make = true;
				if (make)
					try {
						ring.make_back(size+head);
					}
					catch (std::string& error) {
						daq_state |= OVERRUN;
						//store message in static error, which will be sent to
						//master, the next time it connects. DAQ then needs, reset.
					}
				daq_state |= READING;
				body = 2;
			}
		}
#if PICONF_DAQ_AXI_LOOP
		while (daq_state & READING) {
			zdx_box["daq"]->read((uint32_t*)data);
			size_t what = ((uint32_t*)data)[0];
			if (what & (1 << 0)) {//also send what as header to master?!
				daq_state |= OVERRUN;
				return;
			}
#else
		if (daq_state & READING) {
#endif
			size_t copy = size;
			if (size > data_size)
				copy = data_size;
			size -= copy;
			if (body == 2) {
				((uint32_t*)i_item->data)[0] = htonl(i_item->size);
				((uint32_t*)i_item->data)[1] = htonl(daq_state);
				((uint32_t*)i_item->data)[2] = htonl(ring.item_used());
				((uint32_t*)i_item->data)[3] = htonl(ring.data_used());
			}
			for (size_t i = body == 2 ? head : 0; i < copy; ++i)
				i_item->data[i] = ((uint32_t*)data)[i+body];
			if (!size) {
				daq_state &= ~READING;
				++i_item;
				if (i_item_done == item_sync) {
					i_item_done = 1;
					if (daq_state & STALLED)
						daq_state |= OVERRUN;
					else
						daq_state |= STALLED;
				}
				else
					++i_item_done;
			}
			body = 1;
		}
		if (!(daq_state & STALLED)) {
			r_item_done = i_item_done;
			r_item = i_item;
		}
		if (!(io and daq_state & READOUT))
			return;
		if (!io->hasToRead()) {
			io->fetch(&o_item_done, sizeof o_item_done);
			o_item_done = ntohl(o_item_done);
		}
		size_t item_used = ring.item_used(), item_left = r_item_done-o_item_done;
		if (daq_state & READING)
			--item_used;
		for (size_t i = 0; i < item_used-item_left; ++i)
			ring.pop_front();
		if (!item_left)
			daq_state &= ~STALLED;
		if (daq_state & SYNCING) {
			daq_state &= ~SYNCING;
			o_item = ring.begin();
		}
		if (daq_state & WRITING) {
			if (!io->hasToWrite()) {
				daq_state &= ~WRITING;
				++o_item;
			}
		}
		else {
			if (o_item == ring.begin()) {
				try {
					ring.fuse_front(r_item);
				}
				catch (std::string& error) {
				}
				o_item = ring.begin();
			}
		}
		if (!(daq_state & WRITING) and ring.item_used()) {
			if (o_item != r_item) {
				daq_state |= WRITING;
				io->push(o_item->data, o_item->size);
			}
		}
	}
}

bool piconf::daq_tag::process() {
	return data_tag::process();
	if (!(state & RESPONDING)) {
		if (daq_state & INVALID) {
			*log << time() << "DAQ tag invalid" << std::endl;
			return false;
		}
	}
	return true;
}

void piconf::daq_tag::destruct() {
	delete[] data;
	data = 0;
}

#endif
