/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_FILE_TAG_
#define _PICONF_FILE_TAG_

#include "../piconfs.hpp"
#include "file_tag.hpp"
#include <string>
#include <string.h>

piconf::file_tag::file_tag(std::ofstream& log, pugi::xml_node tpl, pugi::xml_node tag, VIO::IO* io)
: data_tag(log, tag, io) {
	this->tpl = tpl;
	file_buff = 0;
	file_state = 0;
	file.chunk(4096, 4096);
	int code = 0;
	pugi::xml_attribute type = tag.attribute("type");
	pugi::xml_node path = tpl.child(type.value());
	if (!path) {
		tag.text().set(((std::string)"File type '"+type.value()+"' unknown").c_str());
		code |= error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
		file_state |= INVALID;
	}
	pugi::xml_attribute readback = tag.attribute("readback");
	if (readback)
		file_state |= READBACK;
	if (!(file_state & INVALID)) {
		if (file_state & READBACK) {
			if (!strlen(tag.attribute("readback").value())) {
				if (strcmp(path.attribute("content").value(), "text")) {
					tag.text().set("Not a text file");
					code |= error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
					file_state |= INVALID;
				}
			}
		}
		else {
			file_size = tag.attribute("size").as_uint();
			if (!file_size) {
				tag.text().set("File size of 0 does not make sense");
				code |= error_t::XML_TAG_ERROR | error_t::PICONFM_ERROR;
				file_state |= INVALID;
			}
			else
			if (file_size > MAX_FILE_SIZE) {
				tag.text().set(((std::string)"File size of "+calculate_size(file_size)+" exceeds maximum of "+calculate_size(MAX_FILE_SIZE)).c_str());
				code |= error_t::MAXIMUM_ERROR | error_t::PICONFM_ERROR;
				file_state |= INVALID;
			}
		}
	}
	if (!(file_state & INVALID)) {
		file.assign(path.text().get());
		if (file_state & READBACK) {
			file.open();
			file_size = file.size();
			tag.append_attribute("size") = to_string(file_size).c_str();
		}
		try {
			file_buff = new char[file_size];
		}
		catch (std::bad_alloc& error) {
			std::string text = "No RAM left to reserve file of "+calculate_size(file_size);
			log << time() << text << std::endl;
			tag.text().set(text.c_str());
			code |= error_t::MAXIMUM_ERROR | error_t::PICONFS_ERROR;
			file_state |= INVALID;
		}
	}
	if (!(file_state & INVALID)) {
		tag.text().set("");
	}
	else {
		if (!tag.attribute("error"))
			tag.append_attribute("error");
		tag.attribute("error").set_value(to_string(code).c_str());
	}
	respond();
	log << time() << "Parsed file_tag" << std::endl;
}

bool piconf::file_tag::process() {
	data_tag::process();
	if (!(state & RESPONDING)) {
		if (file_state & INVALID) {
			*log << time() << "File invalid" << std::endl;
			return false;
		}
		if (file_state & READBACK) {
			if (file_state & WRITING_FILE) {
				if (!io->hasToWrite()) {
					file_state |= DONE;
					*log << time() << "Sending file done" << std::endl;
				}
			}
			else {
				if (file_state & READING_FILE) {
					if (!file.hasToRead()) {
						file_state |= WRITING_FILE;
						io->push(file_buff, file_size);
						*log << time() << "Loading file done, sending file" << std::endl;
					}
				}
				else {
					file.fetch(file_buff, file_size);
					file_state |= READING_FILE;
					*log << time() << "Loading file" << std::endl;
				}

			}
		}
		else {
			if (file_state & WRITING_FILE) {
				if (!file.hasToWrite()) {
					file_state |= DONE;
					*log << time() << "Writing file done" << std::endl;
				}
			}
			else {
				if (file_state & READING_FILE) {
					if (!io->hasToRead()) {
						file_state |= WRITING_FILE;
						try {
							file.remove();
						}
						catch (VIO::IO::Error& error) {
						}
						try {
							file.open(O_RDWR | O_CREAT);
							file.push(file_buff, file_size);
							*log << time() << "Receiving file done, writing file" << std::endl;
						}
						catch (VIO::IO::Error& error) {
							file_state |= DONE;
							*log << time() << "Error: " << error.what() << std::endl;
							return false;
						}
					}
				}
				else {
					file_state |= READING_FILE;
					io->fetch(file_buff, file_size);
					*log << time() << "Receiving file" << std::endl;
				}
			}

		}
	}
	if (file_state & DONE) {
		return false;
	}
	return true;
}

piconf::file_tag::~file_tag() {
	delete[] file_buff;
}

#endif
