/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _PICONF_IO_TAG_
#define _PICONF_IO_TAG_

#include "io_tag.hpp"
#include "../piconfs.hpp"
#include "../../../ape/ape.hpp"
#include "conf_tag.hpp"
#include "file_tag.hpp"
#include "shell_tag.hpp"
#include "daq_tag.hpp"
#include <string.h>
#include <arpa/inet.h>
#include <iostream>
#include <cstdio>
#include <signal.h>

std::ofstream piconf::io_tag::log;

piconf::io_tag::io_tag(const char* tpl_file)
: tag() {
	state = 0;
	tag = 0;
	xml = 0;
	size = 0;
	tpl = tpl_tree.parse(tpl_file, PICONF_ROOT);
	pugi::xml_attribute logfile = tpl.child("io").attribute("logfile");
	if (logfile) {
		log.open(logfile.value(), std::ios_base::app | std::ios_base::out);
		if (!log.is_open())
			throw (std::string)"Cannot open logfile '"+logfile.value()+"'";
	}
	else
		log.open("/dev/null", std::ios_base::out);
	struct sigaction action;
	action.sa_handler = sigterm;
	action.sa_flags = 0;
	if (-1 == sigemptyset(&action.sa_mask) or -1 == sigaction(SIGTERM, &action, NULL))
		throw (std::string)"Cannot set signal handler";
	activate.tv_nsec = 0;
	activate.tv_sec = tpl.child("io").attribute("activate").as_uint();
	if (!activate.tv_sec)
		activate.tv_sec = piconf::ACTIVATE;
	timeout.tv_nsec = 0;
	timeout.tv_sec = tpl.child("io").attribute("timeout").as_uint();
	if (!timeout.tv_sec)
		timeout.tv_sec = piconf::TIMEOUT;
	ape::inherit_attributes(tpl.child("conf"));
	log << time() << "Started piconf" << std::endl;
	conf_tag::construct(log, tpl.child("conf"));
	data_tag::hooked_t* hooked;
	daq_tag::construct(log, tpl.child("daq"), hooked, activate);
	this->hooked.push_back(hooked);
	ape::tree conf;
	std::string conf_file = tpl.child("io").attribute("defconf").value();
	conf_tag(log, tpl.child("conf"), conf.parse(conf_file.c_str(), PICONF_ROOT).child("conf"), 0);
}

bool piconf::io_tag::process() {
	try {
		if (queued.size()) {
			if (state & KNOWING_SIZE) {
				if (state & KNOWING_XML) {
					if (state & PROCESSING) {
						if (!tag->process()) {
							if (tag->hook_me_up()) {
								log << time() << "Hooking up tag" << std::endl;
								state |= HOOKING_UP;
							}
							else {
								state = 0;
								log << time() << "Processing of tag done" << std::endl;
							}
							return true;
						}
					}
					else {
						pugi::xml_node root;
						try {
							root = tag_tree.parse(xml, 0, false);
						}
						catch (std::string& error) {
							state = CLOSE;
							log << time() << "Failed parsing tag: " << error << std::endl;
							return true;
						}
						if (!strcmp(root.name(), "conf")) {
							delete tag;
							tag = new conf_tag(log, tpl.child("conf"), root, queued.front());
						}
						else
						if (!strcmp(root.name(), "file")) {
							delete tag;
							tag = new file_tag(log, tpl.child("file"), root, queued.front());
						}
						else
						if (!strcmp(root.name(), "shell")) {
							delete tag;
							tag = new shell_tag(log, tpl.child("shell"), root, queued.front());
						}
						else
						if (!strcmp(root.name(), "daq")) {
							delete tag;
							tag = new daq_tag(log, tpl.child("daq"), root, queued.front());
						}
						else {
							state = CLOSE;
							log << time() << "Received unknown tag '" << root.name() << "'" << std::endl;
							return true;
						}
						state |= PROCESSING;
						return true;
					}
				}
				else {
					if (state & READING_XML) {
						if (!queued.front()->hasToRead()) {
							if (xml[size-1] != '\0') {
								state = CLOSE;
								log << time() << "Received non null terminated XML" << std::endl;
								return true;
							}
							state |= KNOWING_XML;
							return true;
						}
					}
					else {
						state |= READING_XML;
						delete[] xml;
						xml = new char[size];
						queued.front()->fetch(xml, size);
					}
				}
			}
			else {
				if (state & READING_SIZE) {
					if (!queued.front()->hasToRead()) {
						size = ntohl(size);
						if (!size)
							state = 0;
						if (size > MAX_XML_SIZE) {
							state = CLOSE;
							log << time() << "XML size of " << calculate_size(size) << " exceeds maximum of " << calculate_size(MAX_XML_SIZE) << std::endl;
							return true;
						}
						state |= KNOWING_SIZE;
						return true;
					}
				}
				else {
					state |= READING_SIZE;
					queued.front()->fetch(&size, sizeof size);
				}
			}
		}
		for (hooked_t::iterator hook = hooked.begin(); hook != hooked.end(); ++hook)
			(*hook)->process();
		data_tag::hooked_t::ios_t queue = data_tag::hook_me_down();
		queued.insert(queued.end(), queue.begin(), queue.end());
		queue.clear();
		VIO::IO::process();
	}
	catch (VIO::IO::Error& error) {
		if (error.instance) {
			if (error.instance == VIO::IO::timer)
				throw error;
			if (queued.size() and error.instance == queued.front())
				handle(error);
			else
				handle(error, (VIO::IO*)error.instance);
		}
		else {
			delete tag;
			tag = 0;
			delete[] xml;
			xml = 0;
			log << time() << "Aborting, fatal error: " << error.what() << std::endl;
			throw error;
		}

	}
	return true;
}

void piconf::io_tag::sigterm(int signum) {
	log << time() << "Received SIGTERM, terminating." << std::endl;
	exit(error.get());
}

piconf::io_tag::~io_tag() {
	daq_tag::destruct();
	conf_tag::destruct();
}

#endif
