/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _AXI_
#define _AXI_

#include "AXI.hpp"

#include <inttypes.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <string>
#include <iostream>
#include <exception>

VIO::AXI::Error::Error(const char* context, const char* message, const void* instance, int code) {
	this->context = context;
	if (message)
		this->context += (std::string)": "+message;
	this->instance = instance;
	this->code = code;
}

const char* VIO::AXI::Error::what() const throw() {
	return context.c_str();
}

VIO::AXI::Error::~Error() throw() {
}

VIO::AXI::AXI() {
	construct();
	fd = -1;
	next(0);
}

VIO::AXI::AXI(off_t base_addr, AXI* hsi_axi, AXI* hso_axi, uint8_t done_bit, uint8_t idle_bit) {
	construct();
	open(base_addr, hsi_axi, hso_axi, done_bit, idle_bit);
	next(0);
}

void VIO::AXI::open(off_t base_addr, AXI* hsi_axi, AXI* hso_axi, uint8_t done_bit, uint8_t idle_bit) {
	hsi = hsi_axi;
	hso = hso_axi;
	done = 1 << done_bit;
	idle = 1 << idle_bit;
	fd = ::open(device.c_str(), O_RDWR | O_SYNC);
	if (fd < 0)
		throw Error(((std::string)"Cannot open device '"+device.c_str()+"'").c_str(), strerror(errno), this, 1);
	data = (volatile uint32_t*)mmap(0, getpagesize(), PROT_READ | PROT_WRITE, MAP_SHARED, fd, base_addr);
	if (data < 0)
		throw Error("Cannot map memory", strerror(errno), this, 2);
}

void VIO::AXI::next(AXI* nxt) {
	this->nxt = nxt;
}

void VIO::AXI::read(uint32_t* word) {
	if (hsi and hso) {
		wait_until(done);
		*word = *data;
		if (nxt)
			nxt->read(word+1);
		set(idle);
		wait_until_not(done);
		unset(idle);
	}
	else {
		*word = *data;
		if (nxt)
			nxt->read(word+1);
	}
}

void VIO::AXI::write(const uint32_t* word) {
	*data = *word;
	if (nxt)
		nxt->write(word+1);
	if (hsi and hso) {
		set(done);
		wait_until(idle);
		unset(done);
		wait_until_not(idle);//move to top?
	}
}

void VIO::AXI::close() {
	::close(fd);
}

VIO::AXI::~AXI() {
	close();
}

#endif
