/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2013  Rouven Spreckels  <nevuor@piconf.org>                 *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _VIO_FILE_
#define _VIO_FILE_

#include "File.hpp"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string>
#include <string.h>
#include <cstdio>

VIO::File::Error::Error(const char* context, const char* message, const void* instance, int code)
: IO::Error(context, message, instance, code) {
}

VIO::File::File(const char* file)
: IO() {
	state = 0;
	if (file)
		assign(file);
}

void VIO::File::assign(const char* file) {
	this->file = file;
}

void VIO::File::open(int flags, mode_t mode) {
	fd = ::open(file.c_str(), flags | O_NONBLOCK, mode);
	if (fd < 0)
		throw Error(((std::string)"Cannot open file '"+file+"'").c_str(), strerror(errno), this, OPEN);
	state |= OPENED;
}

bool VIO::File::isOpen() {
	return state & OPENED;
}

bool VIO::File::exists() {
	return exists(file.c_str());
}

bool VIO::File::exists(const char* file) {
	struct stat data;
	return !stat(file, &data);
}

void VIO::File::remove() {
	remove(file.c_str());
}

void VIO::File::remove(const char* file) {
	if (0 != ::remove(file))
		throw Error(((std::string)"Cannot remove file '"+file+"'").c_str(), strerror(errno), 0, REMOVE);
}

std::string VIO::File::basename(std::string file) {
	return file.substr(file.find_last_of(FILE_PATH_DELIMITER)+1);
}

std::string VIO::File::dirname(std::string file) {
	return file.substr(0, file.find_last_of(FILE_PATH_DELIMITER));
}

void VIO::File::mkdir() {
	mkdir(dirname(file).c_str());
}

void VIO::File::mkdir(const char* dir) {
	if (-1 == ::mkdir(dir, 0777))
		throw Error(((std::string)"Cannot create directory '"+dir+"'").c_str(), strerror(errno), 0, MKDIR);
}

size_t VIO::File::size() {
	struct stat stat_buf;
	ssize_t rc = fstat(fd, &stat_buf);
	if (rc < 0)
		throw Error("Cannot stat file", strerror(errno), this, FSTAT);
	return stat_buf.st_size;
}

size_t VIO::File::size(const char* file) {
	struct stat stat_buf;
	ssize_t rc = stat(file, &stat_buf);
	if (rc < 0)
		throw Error(((std::string)"Cannot stat file '"+file+"'").c_str(), strerror(errno), 0, STAT);
	return stat_buf.st_size;
}

void VIO::File::clear() {
	IO::clear();
	state = 0;
}

#endif
