/******************************************************************************
 *  This file is part of piconf.                                              *
 *                                                                            *
 *  Copyright (C) 2012, 2013  Rouven Spreckels  <nevuor@piconf.org>           *
 *                                                                            *
 *  piconf is free software: you can redistribute it and/or modify            *
 *  it under the terms of the GNU Affero General Public License version 3 as  *
 *  published by the Free Software Foundation on 19 November 2007.            *
 *                                                                            *
 *  piconf is distributed in the hope that it will be useful,                 *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU Affero General Public License for more details.                       *
 *                                                                            *
 *  You should have received a copy of the GNU Affero General Public License  *
 *  along with piconf.  If not, see <http://www.gnu.org/licenses/>.           *
 ******************************************************************************/

#ifndef _VIO_IO_
#define _VIO_IO_

#include "IO.hpp"

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <set>
#include <vector>
#include <exception>
#include <string>
#include <sstream>
#include <signal.h>

VIO::IO::IOS VIO::IO::ios;

VIO::IO::IOS* VIO::IO::iox = &VIO::IO::ios;

VIO::IO* VIO::IO::timer = 0;

void VIO::IO::initialize() {
	static bool done = false;
	if (!done) {
		done = true;
		struct sigaction action;
		action.sa_handler = SIG_IGN;
		action.sa_flags = 0;
		if (-1 == sigemptyset(&action.sa_mask) or -1 == sigaction(SIGPIPE, &action, NULL))
			throw Error("Cannot set signal handler", strerror(errno), 0, SIGNAL);
	}
}

VIO::IO::Data::Data(const void* data, size_t size, bool copy) {
	this->copy = 0;
	assign(data, size, copy);
}

VIO::IO::Data::Data(const Data& that) {
	copy = 0;
	assign(that.data, that.size, that.copy);
}

VIO::IO::Data& VIO::IO::Data::operator=(const Data& that) {
	assign(that.data, that.size, copy);
	left = that.left;
	return *this;
}

void VIO::IO::Data::assign(const void* data, size_t size, bool copy) {
	delete[] this->copy;
	if (copy)
		memcpy(this->copy = this->data = new char[size], data, size);
	else
		this->data = (char*)data;
	left = this->size = size;
}

VIO::IO::Data::~Data() {
	delete[] copy;
}

VIO::IO::IO(const void* that) {
	initialize();
	this->that = that ? that : this;
	setTimer();
	last.tv_sec = 0;
	last.tv_nsec = 0;
	fd = -1;
	select = 0;
	fetched.size = 0;
	fetched.left = 0;
	pushed.size = 0;
	pushed.left = 0;
	readChunk = 0;
	writeChunk = 0;
	ios.insert(this);
}

void VIO::IO::chunk(size_t readChunk, size_t writeChunk) {
	this->readChunk = readChunk;
	this->writeChunk = writeChunk;
}

void VIO::IO::configure() {
	int flags;
	if (-1 == (flags = fcntl(fd, F_GETFL, 0)))
		flags = 0;
	if (-1 == fcntl(fd, F_SETFL, flags | O_NONBLOCK))
		throw Error("Cannot set socket to non-blocking", strerror(errno), that, SETSOCKOPT);
}

void VIO::IO::readable() {
	if (!fetched.left)
		read();
	if (!destinations.front().size)
		terminate(destinations.front().left);
	if (fetched.left) {
		if (!destinations.front().size) {
			memcpy(destinations.front().data, fetched.data-(fetched.size-fetched.left), fetched.size-fetched.left);
		}
		else {
			if (fetched.left < destinations.front().left) {
				memcpy(destinations.front().data, fetched.data, fetched.left);
				fetched.size = fetched.left = 0;
			}
			else {
				memcpy(destinations.front().data, fetched.data, destinations.front().left);
				fetched.size = destinations.front().left;
				fetched.left -= destinations.front().left;
				fetched.data += destinations.front().left;
			}
		}
		if (!fetched.left)
			fetched.assign(0, 0, false);
	}
	destinations.front().left -= fetched.size;
	destinations.front().data += fetched.size;
	if (!destinations.front().left) {
		destinations.erase(destinations.begin());
		if (!destinations.size())
			select &= ~READABLE;
	}
	if (fetched.left) {
		select &= ~READABLE;
		if (destinations.size())
			VIO::IO::readable();
		else
			fetched.assign(fetched.data, fetched.left, true);
	}
	else {
		if (destinations.size())
			select |= READABLE;
	}
}

void VIO::IO::writable() {
	write();
	sources.front().left -= pushed.size;
	sources.front().data += pushed.size;
	if (!sources.front().left) {
		sources.erase(sources.begin());
		if (!sources.size())
			select &= ~WRITABLE;
	}
}

void VIO::IO::especial() {
	throw Error("Something espacial happend", 0, that, ESPECIAL);
}

void VIO::IO::activity() {
	last = *time;
}

void VIO::IO::timewise() {
}

void VIO::IO::read() {
	size_t size = destinations.front().left;
	if (readChunk and readChunk < size)
		size = readChunk;
	ssize_t bytes = ::read(fd, destinations.front().data, size);
	if (0 >= bytes)
		throw Error("Cannot read", strerror(errno), that, READ);
	fetched.data = destinations.front().data;
	fetched.size = bytes;
	fetched.left = 0;
}

void VIO::IO::write() {
	size_t size = sources.front().left;
	if (writeChunk and writeChunk < size)
		size = writeChunk;
	ssize_t bytes = ::write(fd, sources.front().data, size);
	if (0 >= bytes)
		throw Error("Cannot write", strerror(errno), that, WRITE);
	pushed.data = sources.front().data;
	pushed.size = bytes;
	pushed.left = 0;
}

void VIO::IO::terminate(size_t& left) {
	size_t byte;
	for (byte = 0; byte < fetched.size and fetched.data[byte]; ++byte);
	if (byte == fetched.size) {
		if (left == fetched.size)
			throw Error("String buffer overflow", 0, that, BUFFER);
	}
	else {
		left = byte+1;
		fetched.left = fetched.size-left;
		fetched.size = left;
	}
	fetched.data += fetched.size;
}

VIO::IO::Error::Error(const char* context, const char* message, const void* instance, int code) {
	this->context = context;
	if (message)
		this->context += (std::string)": "+message;
	this->instance = instance;
	this->code = code;
}

const char* VIO::IO::Error::what() const throw() {
	return context.c_str();
}

VIO::IO::Error::~Error() throw() {
}

VIO::IO::Timeout::Timeout(int sec, int usec) {
	this->sec = sec;
	this->usec = usec;
}

const char* VIO::IO::Timeout::what() const throw() {
	return ((std::string)
		"Timeout after "+
		((std::ostringstream&)(std::ostringstream() << sec)).str()+"s"+
		(usec ? " "+((std::ostringstream&)(std::ostringstream() << usec)).str()+"us" : "")
	).c_str();
}

VIO::IO::IO(const IO& io) {
	that = io.that == &io ? this : io.that;
	if (io.time == &io.last)
		setTimer();
	else
		time = io.time;
	last = io.last;
	fd = io.fd;
	select = io.select;
	fetched = io.fetched;
	pushed = io.pushed;
	readChunk = io.readChunk;
	writeChunk = io.writeChunk;
	destinations = io.destinations;
	sources = io.sources;
	ios.erase((IO*)&io);
	ios.insert(this);
}

VIO::IO& VIO::IO::operator=(const IO& io) {
	that = io.that == &io ? this : io.that;
	if (io.time == &io.last)
		setTimer();
	else
		time = io.time;
	last = io.last;
	fd = io.fd;
	select = io.select;
	fetched = io.fetched;
	pushed = io.pushed;
	readChunk = io.readChunk;
	writeChunk = io.writeChunk;
	destinations = io.destinations;
	sources = io.sources;
	ios.erase((IO*)&io);
	return *this;
}

VIO::IO::~IO() {
	IOS::iterator io;
	if ((io = ios.find(this)) != ios.end()) {
		ios.erase(io);
		::close(fd);
	}
}

void VIO::IO::owner(const void* that) {
	this->that = that ? that : this;
}

bool VIO::IO::isValid(int fd) {
	return -1 != fcntl(fd, F_GETFL) or EBADF != errno;
}

bool VIO::IO::isValid() {
	return isValid(this->fd);
}

void VIO::IO::mask(IOS* include, IOS* exclude) {
	iox = include ? include : &ios;
	if (exclude) {
		static IOS box;
		for (IOS::iterator io = iox->begin(); io != iox->end(); ++io)
			if (exclude->find(*io) == exclude->end())
				box.insert(*io);
		iox = &box;
	}
}

bool VIO::IO::process(int sec, int usec) {
	timeval tv, *wait = 0;
	if (sec > -1 and !(sec == 0 and usec == 0)) {
		wait = &tv;
		wait->tv_sec  = sec;
		wait->tv_usec = usec;
	}
	fd_set readable, writable, especial;
	int max = -1;
	FD_ZERO(&readable);
	FD_ZERO(&writable);
	FD_ZERO(&especial);
	for (IOS::iterator io = iox->begin(); io != iox->end(); ++io) {
		if (-1 < (*io)->fd and (*io)->select & (READABLE | WRITABLE | ESPECIAL)) {
			if ((*io)->select & READABLE)
				FD_SET((*io)->fd, &readable);
			if ((*io)->select & WRITABLE)
				FD_SET((*io)->fd, &writable);
			if ((*io)->select & ESPECIAL)
				FD_SET((*io)->fd, &especial);
			if (max < (*io)->fd)
				max = (*io)->fd;
		}
	}
	if (-1 == max)
		return false;
	switch (::select(max+1, &readable, &writable, &especial, wait)) {
		case -1:
			if (errno != EINTR)
				throw Error("Cannot select", strerror(errno), 0, SELECT);
		break;
		case 0:
			throw Timeout(sec, usec);
		break;
	}
	for (IOS::iterator io = iox->begin(); io != iox->end(); ++io) {
		if (-1 < (*io)->fd and !((*io)->select & SLEEPING)) {
			char activity = false;
			if ((*io)->select & (READABLE | WRITABLE | ESPECIAL)) {
				if (FD_ISSET((*io)->fd, &readable)) {
					activity = true;
					(*io)->readable();
				}
				if (FD_ISSET((*io)->fd, &writable)) {
					activity = true;
					(*io)->writable();
				}
				if (FD_ISSET((*io)->fd, &especial)) {
					activity = true;
					(*io)->especial();
				}
			}
			if (activity and (*io)->select & ACTIVITY)
				(*io)->activity();
		}
	}
	for (IOS::iterator io = iox->begin(); io != iox->end(); ++io)
		if ((*io)->select & TIMEWISE)
			(*io)->timewise();
	return true;
}

void VIO::IO::sleep(bool what) {
	if (what)
		select |= SLEEPING;
	else
		select &= ~SLEEPING;
}

void VIO::IO::readability(bool what) {
	if (what)
		select |= READABLE;
	else
		select &= ~READABLE;
}

void VIO::IO::writability(bool what) {
	if (what)
		select |= WRITABLE;
	else
		select &= ~WRITABLE;
}

void VIO::IO::fetch(void* data, size_t size, bool string) {
	destinations.push_back(Data(data, size));
	if (string)
		destinations.back().size = 0;
	if (fetched.left)
		VIO::IO::readable();
	else
		select |= READABLE;
}

void VIO::IO::push(const void* data, size_t size, bool copy) {
	if (!size)
		size = strlen((const char*)data)+1;
	sources.push_back(Data(data, size, copy));
	select |= WRITABLE;
}

size_t VIO::IO::hasToRead(size_t size) {
	size_t bytes = 0;
	for (Datas::iterator destination = destinations.begin(); destination != destinations.end(); ++destination)
		if (!size or size == destination->size)
			bytes += destination->left;
	return bytes;
}

size_t VIO::IO::hasToWrite(size_t size) {
	size_t bytes = 0;
	for (Datas::iterator source = sources.begin(); source != sources.end(); ++source)
		if (!size or size == source->size)
			bytes += source->left;
	return bytes;
}

timespec VIO::IO::diff(timespec start, timespec end) {
	timespec temp;
	if ((end.tv_nsec-start.tv_nsec) < 0) {
		temp.tv_sec = end.tv_sec-start.tv_sec-1;
		temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
	}
	else {
		temp.tv_sec = end.tv_sec-start.tv_sec;
		temp.tv_nsec = end.tv_nsec-start.tv_nsec;
	}
	return temp;
}

char VIO::IO::more(timespec more, timespec less) {
	if (more.tv_sec > less.tv_sec)
		return true;
	else
	if (more.tv_sec == less.tv_sec) {
		if (more.tv_nsec >= less.tv_nsec)
			return true;
	}
	return false;
}

void VIO::IO::setTimer(IO* timer) {
	if (timer)
		time = &timer->last;
	else {
		if (IO::timer)
			time = &IO::timer->last;
		else
			time = &last;
	}
}

timespec VIO::IO::getTime() {
	return *time;
}

timespec VIO::IO::getLastTime() {
	return last;
}

bool VIO::IO::isActive(timespec span) {
	return (!last.tv_sec and !last.tv_nsec) or more(span, diff(last, *time));
}

bool VIO::IO::activate(timespec span) {
	if (!IO::isActive(span)) {
		static timespec mark = {0, 0};
		if (!more(mark, getTime())) {
			mark = getTime();
			return true;
		}
	}
	return false;
}

bool VIO::IO::schedule() {
	return select & SCHEDULE;
}

void VIO::IO::wait(timespec& time) {
	time = *this->time;
}

bool VIO::IO::hasToWait(timespec time, timespec span) {
	return more(span, diff(time, *this->time));
}

void VIO::IO::close() {
	if (-1 == ::close(fd));
		//throw Error("Cannot close", strerror(errno), that, CLOSE);
}

void VIO::IO::clear() {
	select = 0;
	if (destinations.size())
		destinations.erase(destinations.begin());
	if (sources.size())
		sources.erase(sources.begin());
}

void VIO::IO::swap(IO& io) {
	sources.swap(io.sources);
}

#endif
